# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2018.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test Qiskit's inverse gate operation."""

import unittest
import numpy as np

from qiskit import QuantumRegister, ClassicalRegister, QuantumCircuit, pulse
from qiskit.circuit import Clbit
from qiskit.circuit.classical import expr, types
from qiskit.circuit.library import RXGate, RYGate, GlobalPhaseGate
from qiskit.circuit.exceptions import CircuitError
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class TestCircuitProperties(QiskitTestCase):
    """QuantumCircuit properties tests."""

    def test_qarg_numpy_int(self):
        """Test castable to integer args for QuantumCircuit."""
        n = np.int64(12)
        qc1 = QuantumCircuit(n)
        self.assertEqual(qc1.num_qubits, 12)
        self.assertEqual(type(qc1), QuantumCircuit)

    def test_carg_numpy_int(self):
        """Test castable to integer cargs for QuantumCircuit."""
        n = np.int64(12)
        c1 = ClassicalRegister(n)
        qc1 = QuantumCircuit(c1)
        c_regs = qc1.cregs
        self.assertEqual(c_regs[0], c1)
        self.assertEqual(type(qc1), QuantumCircuit)

    def test_carg_numpy_int_2(self):
        """Test castable to integer cargs for QuantumCircuit."""
        qc1 = QuantumCircuit(12, np.int64(12))
        self.assertEqual(len(qc1.clbits), 12)
        self.assertTrue(all(isinstance(bit, Clbit) for bit in qc1.clbits))
        self.assertEqual(type(qc1), QuantumCircuit)

    def test_qarg_numpy_int_exception(self):
        """Test attempt to pass non-castable arg to QuantumCircuit."""
        self.assertRaises(CircuitError, QuantumCircuit, "string")

    def test_warning_on_noninteger_float(self):
        """Test warning when passing non-integer float to QuantumCircuit"""
        self.assertRaises(CircuitError, QuantumCircuit, 2.2)
        # but an integer float should pass
        qc = QuantumCircuit(2.0)
        self.assertEqual(qc.num_qubits, 2)

    def test_circuit_depth_empty(self):
        """Test depth of empty circuity"""
        q = QuantumRegister(5, "q")
        qc = QuantumCircuit(q)
        self.assertEqual(qc.depth(), 0)

    def test_circuit_depth_no_reg(self):
        """Test depth of no register circuits"""
        qc = QuantumCircuit()
        self.assertEqual(qc.depth(), 0)

    def test_circuit_depth_meas_only(self):
        """Test depth of measurement only"""
        q = QuantumRegister(1, "q")
        c = ClassicalRegister(1, "c")
        qc = QuantumCircuit(q, c)
        qc.measure(q, c)
        self.assertEqual(qc.depth(), 1)

    def test_circuit_depth_barrier(self):
        """Make sure barriers do not add to depth"""

        #         ┌───┐                     ░ ┌─┐
        #    q_0: ┤ H ├──■──────────────────░─┤M├────────────
        #         ├───┤┌─┴─┐                ░ └╥┘┌─┐
        #    q_1: ┤ H ├┤ X ├──■─────────────░──╫─┤M├─────────
        #         ├───┤└───┘  │  ┌───┐      ░  ║ └╥┘┌─┐
        #    q_2: ┤ H ├───────┼──┤ X ├──■───░──╫──╫─┤M├──────
        #         ├───┤       │  └─┬─┘┌─┴─┐ ░  ║  ║ └╥┘┌─┐
        #    q_3: ┤ H ├───────┼────┼──┤ X ├─░──╫──╫──╫─┤M├───
        #         ├───┤     ┌─┴─┐  │  └───┘ ░  ║  ║  ║ └╥┘┌─┐
        #    q_4: ┤ H ├─────┤ X ├──■────────░──╫──╫──╫──╫─┤M├
        #         └───┘     └───┘           ░  ║  ║  ║  ║ └╥┘
        #    c: 5/═════════════════════════════╩══╩══╩══╩══╩═
        #                                      0  1  2  3  4
        q = QuantumRegister(5, "q")
        c = ClassicalRegister(5, "c")
        qc = QuantumCircuit(q, c)
        qc.h(q[0])
        qc.h(q[1])
        qc.h(q[2])
        qc.h(q[3])
        qc.h(q[4])
        qc.cx(q[0], q[1])
        qc.cx(q[1], q[4])
        qc.cx(q[4], q[2])
        qc.cx(q[2], q[3])
        qc.barrier(q)
        qc.measure(q, c)
        self.assertEqual(qc.depth(), 6)

    def test_circuit_depth_simple(self):
        """Test depth for simple circuit"""
        #      ┌───┐
        # q_0: ┤ H ├──■────────────────────
        #      └───┘  │            ┌───┐┌─┐
        # q_1: ───────┼────────────┤ X ├┤M├
        #      ┌───┐  │  ┌───┐┌───┐└─┬─┘└╥┘
        # q_2: ┤ X ├──┼──┤ X ├┤ X ├──┼───╫─
        #      └───┘  │  └───┘└───┘  │   ║
        # q_3: ───────┼──────────────┼───╫─
        #           ┌─┴─┐┌───┐       │   ║
        # q_4: ─────┤ X ├┤ X ├───────■───╫─
        #           └───┘└───┘           ║
        # c: 1/══════════════════════════╩═
        #                                0
        q = QuantumRegister(5, "q")
        c = ClassicalRegister(1, "c")
        qc = QuantumCircuit(q, c)
        qc.h(q[0])
        qc.cx(q[0], q[4])
        qc.x(q[2])
        qc.x(q[2])
        qc.x(q[2])
        qc.x(q[4])
        qc.cx(q[4], q[1])
        qc.measure(q[1], c[0])
        self.assertEqual(qc.depth(), 5)

    def test_circuit_depth_multi_reg(self):
        """Test depth for multiple registers"""

        #       ┌───┐
        # q1_0: ┤ H ├──■─────────────────
        #       ├───┤┌─┴─┐
        # q1_1: ┤ H ├┤ X ├──■────────────
        #       ├───┤└───┘  │  ┌───┐
        # q1_2: ┤ H ├───────┼──┤ X ├──■──
        #       ├───┤       │  └─┬─┘┌─┴─┐
        # q2_0: ┤ H ├───────┼────┼──┤ X ├
        #       ├───┤     ┌─┴─┐  │  └───┘
        # q2_1: ┤ H ├─────┤ X ├──■───────
        #       └───┘     └───┘
        q1 = QuantumRegister(3, "q1")
        q2 = QuantumRegister(2, "q2")
        c = ClassicalRegister(5, "c")
        qc = QuantumCircuit(q1, q2, c)
        qc.h(q1[0])
        qc.h(q1[1])
        qc.h(q1[2])
        qc.h(q2[0])
        qc.h(q2[1])
        qc.cx(q1[0], q1[1])
        qc.cx(q1[1], q2[1])
        qc.cx(q2[1], q1[2])
        qc.cx(q1[2], q2[0])
        self.assertEqual(qc.depth(), 5)

    def test_circuit_depth_3q_gate(self):
        """Test depth for 3q gate"""

        #       ┌───┐
        # q1_0: ┤ H ├──■────■─────────────────
        #       ├───┤  │  ┌─┴─┐
        # q1_1: ┤ H ├──┼──┤ X ├──■────────────
        #       ├───┤  │  └───┘  │  ┌───┐
        # q1_2: ┤ H ├──┼─────────┼──┤ X ├──■──
        #       ├───┤┌─┴─┐       │  └─┬─┘┌─┴─┐
        # q2_0: ┤ H ├┤ X ├───────┼────┼──┤ X ├
        #       ├───┤└─┬─┘     ┌─┴─┐  │  └───┘
        # q2_1: ┤ H ├──■───────┤ X ├──■───────
        #       └───┘          └───┘
        q1 = QuantumRegister(3, "q1")
        q2 = QuantumRegister(2, "q2")
        c = ClassicalRegister(5, "c")
        qc = QuantumCircuit(q1, q2, c)
        qc.h(q1[0])
        qc.h(q1[1])
        qc.h(q1[2])
        qc.h(q2[0])
        qc.h(q2[1])
        qc.ccx(q2[1], q1[0], q2[0])
        qc.cx(q1[0], q1[1])
        qc.cx(q1[1], q2[1])
        qc.cx(q2[1], q1[2])
        qc.cx(q1[2], q2[0])
        self.assertEqual(qc.depth(), 6)

    def test_circuit_depth_conditionals1(self):
        """Test circuit depth for conditional gates #1."""

        #      ┌───┐     ┌─┐
        # q_0: ┤ H ├──■──┤M├─────────────────
        #      ├───┤┌─┴─┐└╥┘┌─┐
        # q_1: ┤ H ├┤ X ├─╫─┤M├──────────────
        #      ├───┤└───┘ ║ └╥┘ ┌───┐
        # q_2: ┤ H ├──■───╫──╫──┤ H ├────────
        #      ├───┤┌─┴─┐ ║  ║  └─╥─┘  ┌───┐
        # q_3: ┤ H ├┤ X ├─╫──╫────╫────┤ H ├─
        #      └───┘└───┘ ║  ║    ║    └─╥─┘
        #                 ║  ║ ┌──╨──┐┌──╨──┐
        # c: 4/═══════════╩══╩═╡ 0x2 ╞╡ 0x4 ╞
        #                 0  1 └─────┘└─────┘
        size = 4
        q = QuantumRegister(size, "q")
        c = ClassicalRegister(size, "c")
        qc = QuantumCircuit(q, c)

        qc.h(q[0])
        qc.h(q[1])
        qc.h(q[2])
        qc.h(q[3])
        qc.cx(q[0], q[1])
        qc.cx(q[2], q[3])
        qc.measure(q[0], c[0])
        qc.measure(q[1], c[1])
        with self.assertWarns(DeprecationWarning):
            qc.h(q[2]).c_if(c, 2)
        with self.assertWarns(DeprecationWarning):
            qc.h(q[3]).c_if(c, 4)
        self.assertEqual(qc.depth(), 5)

    def test_circuit_depth_conditionals2(self):
        """Test circuit depth for conditional gates #2."""

        #      ┌───┐     ┌─┐┌─┐
        # q_0: ┤ H ├──■──┤M├┤M├──────────────
        #      ├───┤┌─┴─┐└╥┘└╥┘
        # q_1: ┤ H ├┤ X ├─╫──╫───────────────
        #      ├───┤└───┘ ║  ║  ┌───┐
        # q_2: ┤ H ├──■───╫──╫──┤ H ├────────
        #      ├───┤┌─┴─┐ ║  ║  └─╥─┘  ┌───┐
        # q_3: ┤ H ├┤ X ├─╫──╫────╫────┤ H ├─
        #      └───┘└───┘ ║  ║    ║    └─╥─┘
        #                 ║  ║ ┌──╨──┐┌──╨──┐
        # c: 4/═══════════╩══╩═╡ 0x2 ╞╡ 0x4 ╞
        #                 0  0 └─────┘└─────┘
        size = 4
        q = QuantumRegister(size, "q")
        c = ClassicalRegister(size, "c")
        qc = QuantumCircuit(q, c)

        qc.h(q[0])
        qc.h(q[1])
        qc.h(q[2])
        qc.h(q[3])
        qc.cx(q[0], q[1])
        qc.cx(q[2], q[3])
        qc.measure(q[0], c[0])
        qc.measure(q[0], c[0])
        with self.assertWarns(DeprecationWarning):
            qc.h(q[2]).c_if(c, 2)
        with self.assertWarns(DeprecationWarning):
            qc.h(q[3]).c_if(c, 4)
        self.assertEqual(qc.depth(), 6)

    def test_circuit_depth_conditionals3(self):
        """Test circuit depth for conditional gates #3."""

        #      ┌───┐┌─┐
        # q_0: ┤ H ├┤M├───■────────────
        #      ├───┤└╥┘   │   ┌─┐
        # q_1: ┤ H ├─╫────┼───┤M├──────
        #      ├───┤ ║    │   └╥┘┌─┐
        # q_2: ┤ H ├─╫────┼────╫─┤M├───
        #      ├───┤ ║  ┌─┴─┐  ║ └╥┘┌─┐
        # q_3: ┤ H ├─╫──┤ X ├──╫──╫─┤M├
        #      └───┘ ║  └─╥─┘  ║  ║ └╥┘
        #            ║ ┌──╨──┐ ║  ║  ║
        # c: 4/══════╩═╡ 0x2 ╞═╩══╩══╩═
        #            0 └─────┘ 1  2  3
        size = 4
        q = QuantumRegister(size, "q")
        c = ClassicalRegister(size, "c")
        qc = QuantumCircuit(q, c)

        qc.h(q[0])
        qc.h(q[1])
        qc.h(q[2])
        qc.h(q[3])
        qc.measure(q[0], c[0])
        with self.assertWarns(DeprecationWarning):
            qc.cx(q[0], q[3]).c_if(c, 2)

        qc.measure(q[1], c[1])
        qc.measure(q[2], c[2])
        qc.measure(q[3], c[3])
        self.assertEqual(qc.depth(), 4)

    def test_circuit_depth_bit_conditionals1(self):
        """Test circuit depth for single bit conditional gates #1."""

        #      ┌───┐┌─┐
        # q_0: ┤ H ├┤M├─────────────────────────
        #      ├───┤└╥┘      ┌───┐
        # q_1: ┤ H ├─╫───────┤ H ├──────────────
        #      ├───┤ ║ ┌─┐   └─╥─┘
        # q_2: ┤ H ├─╫─┤M├─────╫────────────────
        #      ├───┤ ║ └╥┘     ║        ┌───┐
        # q_3: ┤ H ├─╫──╫──────╫────────┤ H ├───
        #      └───┘ ║  ║      ║        └─╥─┘
        #            ║  ║ ┌────╨────┐┌────╨────┐
        # c: 4/══════╩══╩═╡ c_0=0x1 ╞╡ c_2=0x0 ╞
        #            0  2 └─────────┘└─────────┘
        size = 4
        q = QuantumRegister(size, "q")
        c = ClassicalRegister(size, "c")
        qc = QuantumCircuit(q, c)

        qc.h(q[0])
        qc.h(q[1])
        qc.h(q[2])
        qc.h(q[3])
        qc.measure(q[0], c[0])
        qc.measure(q[2], c[2])
        with self.assertWarns(DeprecationWarning):
            qc.h(q[1]).c_if(c[0], True)
        with self.assertWarns(DeprecationWarning):
            qc.h(q[3]).c_if(c[2], False)
        self.assertEqual(qc.depth(), 3)

    def test_circuit_depth_bit_conditionals2(self):
        """Test circuit depth for single bit conditional gates #2."""

        #      ┌───┐┌─┐                                                          »
        # q_0: ┤ H ├┤M├──────────────────────────────■─────────────────────■─────»
        #      ├───┤└╥┘      ┌───┐                 ┌─┴─┐                   │     »
        # q_1: ┤ H ├─╫───────┤ H ├─────────────────┤ X ├───────────────────┼─────»
        #      ├───┤ ║ ┌─┐   └─╥─┘                 └─╥─┘                 ┌─┴─┐   »
        # q_2: ┤ H ├─╫─┤M├─────╫─────────────────────╫──────────■────────┤ H ├───»
        #      ├───┤ ║ └╥┘     ║        ┌───┐        ║        ┌─┴─┐      └─╥─┘   »
        # q_3: ┤ H ├─╫──╫──────╫────────┤ H ├────────╫────────┤ X ├────────╫─────»
        #      └───┘ ║  ║      ║        └─╥─┘        ║        └─╥─┘        ║     »
        #            ║  ║ ┌────╨────┐┌────╨────┐┌────╨────┐┌────╨────┐┌────╨────┐»
        # c: 4/══════╩══╩═╡ c_1=0x1 ╞╡ c_3=0x1 ╞╡ c_0=0x0 ╞╡ c_2=0x0 ╞╡ c_1=0x1 ╞»
        #            0  2 └─────────┘└─────────┘└─────────┘└─────────┘└─────────┘»
        # «
        # «q_0: ───────────
        # «
        # «q_1: ─────■─────
        # «          │
        # «q_2: ─────┼─────
        # «        ┌─┴─┐
        # «q_3: ───┤ H ├───
        # «        └─╥─┘
        # «     ┌────╨────┐
        # «c: 4/╡ c_3=0x1 ╞
        # «     └─────────┘
        size = 4
        q = QuantumRegister(size, "q")
        c = ClassicalRegister(size, "c")
        qc = QuantumCircuit(q, c)

        qc.h(q[0])
        qc.h(q[1])
        qc.h(q[2])
        qc.h(q[3])
        qc.measure(q[0], c[0])
        qc.measure(q[2], c[2])
        with self.assertWarns(DeprecationWarning):
            qc.h(q[1]).c_if(c[1], True)
        with self.assertWarns(DeprecationWarning):
            qc.h(q[3]).c_if(c[3], True)
        with self.assertWarns(DeprecationWarning):
            qc.cx(0, 1).c_if(c[0], False)
        with self.assertWarns(DeprecationWarning):
            qc.cx(2, 3).c_if(c[2], False)
        with self.assertWarns(DeprecationWarning):
            qc.ch(0, 2).c_if(c[1], True)
        with self.assertWarns(DeprecationWarning):
            qc.ch(1, 3).c_if(c[3], True)
        self.assertEqual(qc.depth(), 4)

    def test_circuit_depth_bit_conditionals3(self):
        """Test circuit depth for single bit conditional gates #3."""

        #      ┌───┐┌─┐
        # q_0: ┤ H ├┤M├──────────────────────────────────────
        #      ├───┤└╥┘   ┌───┐                     ┌─┐
        # q_1: ┤ H ├─╫────┤ H ├─────────────────────┤M├──────
        #      ├───┤ ║    └─╥─┘    ┌───┐            └╥┘┌─┐
        # q_2: ┤ H ├─╫──────╫──────┤ H ├─────────────╫─┤M├───
        #      ├───┤ ║      ║      └─╥─┘    ┌───┐    ║ └╥┘┌─┐
        # q_3: ┤ H ├─╫──────╫────────╫──────┤ H ├────╫──╫─┤M├
        #      └───┘ ║      ║        ║      └─╥─┘    ║  ║ └╥┘
        #            ║ ┌────╨────┐┌──╨──┐┌────╨────┐ ║  ║  ║
        # c: 4/══════╩═╡ c_0=0x1 ╞╡ 0x2 ╞╡ c_3=0x1 ╞═╩══╩══╩═
        #            0 └─────────┘└─────┘└─────────┘ 1  2  3
        size = 4
        q = QuantumRegister(size, "q")
        c = ClassicalRegister(size, "c")
        qc = QuantumCircuit(q, c)

        qc.h(q[0])
        qc.h(q[1])
        qc.h(q[2])
        qc.h(q[3])
        qc.measure(q[0], c[0])
        with self.assertWarns(DeprecationWarning):
            qc.h(1).c_if(c[0], True)
        with self.assertWarns(DeprecationWarning):
            qc.h(q[2]).c_if(c, 2)
        with self.assertWarns(DeprecationWarning):
            qc.h(3).c_if(c[3], True)
        qc.measure(q[1], c[1])
        qc.measure(q[2], c[2])
        qc.measure(q[3], c[3])
        self.assertEqual(qc.depth(), 6)

    def test_circuit_depth_measurements1(self):
        """Test circuit depth for measurements #1."""

        #      ┌───┐┌─┐
        # q_0: ┤ H ├┤M├─────────
        #      ├───┤└╥┘┌─┐
        # q_1: ┤ H ├─╫─┤M├──────
        #      ├───┤ ║ └╥┘┌─┐
        # q_2: ┤ H ├─╫──╫─┤M├───
        #      ├───┤ ║  ║ └╥┘┌─┐
        # q_3: ┤ H ├─╫──╫──╫─┤M├
        #      └───┘ ║  ║  ║ └╥┘
        # c: 4/══════╩══╩══╩══╩═
        #            0  1  2  3
        size = 4
        q = QuantumRegister(size, "q")
        c = ClassicalRegister(size, "c")
        qc = QuantumCircuit(q, c)

        qc.h(q[0])
        qc.h(q[1])
        qc.h(q[2])
        qc.h(q[3])
        qc.measure(q[0], c[0])
        qc.measure(q[1], c[1])
        qc.measure(q[2], c[2])
        qc.measure(q[3], c[3])
        self.assertEqual(qc.depth(), 2)

    def test_circuit_depth_measurements2(self):
        """Test circuit depth for measurements #2."""

        #      ┌───┐┌─┐┌─┐┌─┐┌─┐
        # q_0: ┤ H ├┤M├┤M├┤M├┤M├
        #      ├───┤└╥┘└╥┘└╥┘└╥┘
        # q_1: ┤ H ├─╫──╫──╫──╫─
        #      ├───┤ ║  ║  ║  ║
        # q_2: ┤ H ├─╫──╫──╫──╫─
        #      ├───┤ ║  ║  ║  ║
        # q_3: ┤ H ├─╫──╫──╫──╫─
        #      └───┘ ║  ║  ║  ║
        # c: 4/══════╩══╩══╩══╩═
        #            0  1  2  3
        size = 4
        q = QuantumRegister(size, "q")
        c = ClassicalRegister(size, "c")
        qc = QuantumCircuit(q, c)

        qc.h(q[0])
        qc.h(q[1])
        qc.h(q[2])
        qc.h(q[3])
        qc.measure(q[0], c[0])
        qc.measure(q[0], c[1])
        qc.measure(q[0], c[2])
        qc.measure(q[0], c[3])
        self.assertEqual(qc.depth(), 5)

    def test_circuit_depth_measurements3(self):
        """Test circuit depth for measurements #3."""

        #      ┌───┐┌─┐
        # q_0: ┤ H ├┤M├─────────
        #      ├───┤└╥┘┌─┐
        # q_1: ┤ H ├─╫─┤M├──────
        #      ├───┤ ║ └╥┘┌─┐
        # q_2: ┤ H ├─╫──╫─┤M├───
        #      ├───┤ ║  ║ └╥┘┌─┐
        # q_3: ┤ H ├─╫──╫──╫─┤M├
        #      └───┘ ║  ║  ║ └╥┘
        # c: 4/══════╩══╩══╩══╩═
        #            0  0  0  0
        size = 4
        q = QuantumRegister(size, "q")
        c = ClassicalRegister(size, "c")
        qc = QuantumCircuit(q, c)

        qc.h(q[0])
        qc.h(q[1])
        qc.h(q[2])
        qc.h(q[3])
        qc.measure(q[0], c[0])
        qc.measure(q[1], c[0])
        qc.measure(q[2], c[0])
        qc.measure(q[3], c[0])
        self.assertEqual(qc.depth(), 5)

    def test_circuit_depth_barriers1(self):
        """Test circuit depth for barriers #1."""

        #      ┌───┐      ░
        # q_0: ┤ H ├──■───░───────────
        #      └───┘┌─┴─┐ ░
        # q_1: ─────┤ X ├─░───────────
        #           └───┘ ░ ┌───┐
        # q_2: ───────────░─┤ H ├──■──
        #                 ░ └───┘┌─┴─┐
        # q_3: ───────────░──────┤ X ├
        #                 ░      └───┘
        q = QuantumRegister(4, "q")
        c = ClassicalRegister(4, "c")
        circ = QuantumCircuit(q, c)
        circ.h(0)
        circ.cx(0, 1)
        circ.barrier(q)
        circ.h(2)
        circ.cx(2, 3)
        self.assertEqual(circ.depth(), 4)

    def test_circuit_depth_barriers2(self):
        """Test circuit depth for barriers #2."""

        #      ┌───┐ ░       ░       ░
        # q_0: ┤ H ├─░───■───░───────░──────
        #      └───┘ ░ ┌─┴─┐ ░       ░
        # q_1: ──────░─┤ X ├─░───────░──────
        #            ░ └───┘ ░ ┌───┐ ░
        # q_2: ──────░───────░─┤ H ├─░───■──
        #            ░       ░ └───┘ ░ ┌─┴─┐
        # q_3: ──────░───────░───────░─┤ X ├
        #            ░       ░       ░ └───┘
        q = QuantumRegister(4, "q")
        c = ClassicalRegister(4, "c")
        circ = QuantumCircuit(q, c)
        circ.h(0)
        circ.barrier(q)
        circ.cx(0, 1)
        circ.barrier(q)
        circ.h(2)
        circ.barrier(q)
        circ.cx(2, 3)
        self.assertEqual(circ.depth(), 4)

    def test_circuit_depth_barriers3(self):
        """Test circuit depth for barriers #3."""

        #      ┌───┐ ░       ░  ░  ░       ░
        # q_0: ┤ H ├─░───■───░──░──░───────░──────
        #      └───┘ ░ ┌─┴─┐ ░  ░  ░       ░
        # q_1: ──────░─┤ X ├─░──░──░───────░──────
        #            ░ └───┘ ░  ░  ░ ┌───┐ ░
        # q_2: ──────░───────░──░──░─┤ H ├─░───■──
        #            ░       ░  ░  ░ └───┘ ░ ┌─┴─┐
        # q_3: ──────░───────░──░──░───────░─┤ X ├
        #            ░       ░  ░  ░       ░ └───┘
        q = QuantumRegister(4, "q")
        c = ClassicalRegister(4, "c")
        circ = QuantumCircuit(q, c)
        circ.h(0)
        circ.barrier(q)
        circ.cx(0, 1)
        circ.barrier(q)
        circ.barrier(q)
        circ.barrier(q)
        circ.h(2)
        circ.barrier(q)
        circ.cx(2, 3)
        self.assertEqual(circ.depth(), 4)

    def test_circuit_depth_2qubit(self):
        """Test finding depth of two-qubit gates only."""

        #      ┌───┐
        # q_0: ┤ H ├──■───────────────────
        #      └───┘┌─┴─┐┌─────────┐   ┌─┐
        # q_1: ─────┤ X ├┤ Rz(0.1) ├─■─┤M├
        #      ┌───┐└───┘└─────────┘ │ └╥┘
        # q_2: ┤ H ├──■──────────────┼──╫─
        #      └───┘┌─┴─┐            │  ║
        # q_3: ─────┤ X ├────────────■──╫─
        #           └───┘               ║
        # c: 1/═════════════════════════╩═
        #                               0
        circ = QuantumCircuit(4, 1)
        circ.h(0)
        circ.cx(0, 1)
        circ.h(2)
        circ.cx(2, 3)
        circ.rz(0.1, 1)
        circ.cz(1, 3)
        circ.measure(1, 0)
        self.assertEqual(circ.depth(lambda x: x.operation.num_qubits == 2), 2)

    def test_circuit_depth_multiqubit_or_conditional(self):
        """Test finding depth of multi-qubit or conditional gates."""

        #      ┌───┐                              ┌───┐
        # q_0: ┤ H ├──■───────────────────────────┤ X ├───
        #      └───┘  │  ┌─────────┐        ┌─┐   └─╥─┘
        # q_1: ───────■──┤ Rz(0.1) ├──────■─┤M├─────╫─────
        #           ┌─┴─┐└──┬───┬──┘      │ └╥┘     ║
        # q_2: ─────┤ X ├───┤ H ├─────■───┼──╫──────╫─────
        #           └───┘   └───┘   ┌─┴─┐ │  ║      ║
        # q_3: ─────────────────────┤ X ├─■──╫──────╫─────
        #                           └───┘    ║ ┌────╨────┐
        # c: 1/══════════════════════════════╩═╡ c_0 = T ╞
        #                                    0 └─────────┘
        circ = QuantumCircuit(4, 1)
        circ.h(0)
        circ.ccx(0, 1, 2)
        circ.h(2)
        circ.cx(2, 3)
        circ.rz(0.1, 1)
        circ.cz(1, 3)
        circ.measure(1, 0)
        with self.assertWarns(DeprecationWarning):
            circ.x(0).c_if(0, 1)
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(
                circ.depth(
                    lambda x: x.operation.num_qubits >= 2 or x.operation.condition is not None
                ),
                4,
            )

    def test_circuit_depth_first_qubit(self):
        """Test finding depth of gates touching q0 only."""

        #      ┌───┐        ┌───┐
        # q_0: ┤ H ├──■─────┤ T ├─────────
        #      └───┘┌─┴─┐┌──┴───┴──┐   ┌─┐
        # q_1: ─────┤ X ├┤ Rz(0.1) ├─■─┤M├
        #      ┌───┐└───┘└─────────┘ │ └╥┘
        # q_2: ┤ H ├──■──────────────┼──╫─
        #      └───┘┌─┴─┐            │  ║
        # q_3: ─────┤ X ├────────────■──╫─
        #           └───┘               ║
        # c: 1/═════════════════════════╩═
        #                               0
        circ = QuantumCircuit(4, 1)
        circ.h(0)
        circ.cx(0, 1)
        circ.t(0)
        circ.h(2)
        circ.cx(2, 3)
        circ.rz(0.1, 1)
        circ.cz(1, 3)
        circ.measure(1, 0)
        self.assertEqual(circ.depth(lambda x: circ.qubits[0] in x.qubits), 3)

    def test_circuit_depth_0_operands(self):
        """Test that the depth can be found even with zero-bit operands."""
        qc = QuantumCircuit(2, 2)
        qc.append(GlobalPhaseGate(0.0), [], [])
        qc.append(GlobalPhaseGate(0.0), [], [])
        qc.append(GlobalPhaseGate(0.0), [], [])
        self.assertEqual(qc.depth(), 0)
        qc.measure([0, 1], [0, 1])
        self.assertEqual(qc.depth(), 1)

    def test_circuit_depth_expr_condition(self):
        """Test that circuit depth respects `Expr` conditions in `IfElseOp`."""
        # Note that the "depth" of control-flow operations is not well defined, so the assertions
        # here are quite weak.  We're mostly aiming to match legacy behaviour of `c_if` for cases
        # where there's a single instruction within the conditional.
        qc = QuantumCircuit(2, 2)
        a = qc.add_input("a", types.Bool())
        with qc.if_test(a):
            qc.x(0)
        with qc.if_test(expr.logic_and(a, qc.clbits[0])):
            qc.x(1)
        self.assertEqual(qc.depth(), 2)
        qc.measure([0, 1], [0, 1])
        self.assertEqual(qc.depth(), 3)

    def test_circuit_depth_expr_store(self):
        """Test that circuit depth respects `Store`."""
        qc = QuantumCircuit(3, 3)
        a = qc.add_input("a", types.Bool())
        qc.h(0)
        qc.cx(0, 1)
        qc.measure([0, 1], [0, 1])
        # Note that `Store` is a "directive", so doesn't increase the depth by default, but does
        # cause qubits 0,1; clbits 0,1 and 'a' to all be depth 3 at this point.
        qc.store(a, qc.clbits[0])
        qc.store(a, expr.logic_and(a, qc.clbits[1]))
        # ... so this use of 'a' should make it depth 4.
        with qc.if_test(a):
            qc.x(2)
        self.assertEqual(qc.depth(), 4)

    def test_circuit_depth_switch(self):
        """Test that circuit depth respects the `target` of `SwitchCaseOp`."""
        qc = QuantumCircuit(QuantumRegister(3, "q"), ClassicalRegister(3, "c"))
        a = qc.add_input("a", types.Uint(3))

        with qc.switch(expr.bit_and(a, qc.cregs[0])) as case:
            with case(case.DEFAULT):
                qc.x(0)
        qc.measure(1, 0)
        self.assertEqual(qc.depth(), 2)

    def test_circuit_size_empty(self):
        """Circuit.size should return 0 for an empty circuit."""
        size = 4
        q = QuantumRegister(size, "q")
        c = ClassicalRegister(size, "c")
        qc = QuantumCircuit(q, c)

        self.assertEqual(qc.size(), 0)

    def test_circuit_size_single_qubit_gates(self):
        """Circuit.size should increment for each added single qubit gate."""
        size = 4
        q = QuantumRegister(size, "q")
        c = ClassicalRegister(size, "c")
        qc = QuantumCircuit(q, c)

        qc.h(q[0])
        self.assertEqual(qc.size(), 1)
        qc.h(q[1])
        self.assertEqual(qc.size(), 2)

    def test_circuit_size_2qubit(self):
        """Circuit.size of only 2-qubit gates."""
        size = 3
        q = QuantumRegister(size, "q")
        c = ClassicalRegister(size, "c")
        qc = QuantumCircuit(q, c)

        qc.cx(q[0], q[1])
        qc.rz(0.1, q[1])
        qc.rzz(0.1, q[1], q[2])
        self.assertEqual(qc.size(lambda x: x.operation.num_qubits == 2), 2)

    def test_circuit_count_ops(self):
        """Test circuit count ops."""
        q = QuantumRegister(6, "q")
        qc = QuantumCircuit(q)
        qc.h(q)
        qc.x(q[1])
        qc.y(q[2:4])
        qc.z(q[3:])
        result = qc.count_ops()

        expected = {"h": 6, "z": 3, "y": 2, "x": 1}

        self.assertIsInstance(result, dict)
        self.assertEqual(expected, result)

    def test_circuit_nonlocal_gates(self):
        """Test num_nonlocal_gates."""

        #      ┌───┐                   ┌────────┐
        # q_0: ┤ H ├───────────────────┤0       ├
        #      ├───┤   ┌───┐           │        │
        # q_1: ┤ H ├───┤ X ├─────────■─┤        ├
        #      ├───┤   └───┘         │ │        │
        # q_2: ┤ H ├─────■───────────X─┤  Iswap ├
        #      ├───┤     │     ┌───┐ │ │        │
        # q_3: ┤ H ├─────┼─────┤ Z ├─X─┤        ├
        #      ├───┤┌────┴────┐├───┤   │        │
        # q_4: ┤ H ├┤ Ry(0.1) ├┤ Z ├───┤1       ├
        #      ├───┤└──┬───┬──┘└───┘   └───╥────┘
        # q_5: ┤ H ├───┤ Z ├───────────────╫─────
        #      └───┘   └───┘            ┌──╨──┐
        # c: 2/═════════════════════════╡ 0x2 ╞══
        #                               └─────┘
        q = QuantumRegister(6, "q")
        c = ClassicalRegister(2, "c")
        qc = QuantumCircuit(q, c)
        qc.h(q)
        qc.x(q[1])
        qc.cry(0.1, q[2], q[4])
        qc.z(q[3:])
        qc.cswap(q[1], q[2], q[3])
        with self.assertWarns(DeprecationWarning):
            qc.iswap(q[0], q[4]).c_if(c, 2)
        result = qc.num_nonlocal_gates()
        expected = 3
        self.assertEqual(expected, result)

    def test_circuit_nonlocal_gates_no_instruction(self):
        """Verify num_nunlocal_gates does not include barriers."""
        # ref: https://github.com/Qiskit/qiskit-terra/issues/4500
        n = 3
        qc = QuantumCircuit(n)
        qc.h(range(n))

        qc.barrier()

        self.assertEqual(qc.num_nonlocal_gates(), 0)

    def test_circuit_connected_components_empty(self):
        """Verify num_connected_components is width for empty"""
        q = QuantumRegister(7, "q")
        qc = QuantumCircuit(q)
        self.assertEqual(7, qc.num_connected_components())

    def test_circuit_connected_components_multi_reg(self):
        """Test tensor factors works over multi registers"""

        #       ┌───┐
        # q1_0: ┤ H ├──■─────────────────
        #       ├───┤┌─┴─┐
        # q1_1: ┤ H ├┤ X ├──■────────────
        #       ├───┤└───┘  │  ┌───┐
        # q1_2: ┤ H ├───────┼──┤ X ├──■──
        #       ├───┤       │  └─┬─┘┌─┴─┐
        # q2_0: ┤ H ├───────┼────┼──┤ X ├
        #       ├───┤     ┌─┴─┐  │  └───┘
        # q2_1: ┤ H ├─────┤ X ├──■───────
        #       └───┘     └───┘
        q1 = QuantumRegister(3, "q1")
        q2 = QuantumRegister(2, "q2")
        qc = QuantumCircuit(q1, q2)
        qc.h(q1[0])
        qc.h(q1[1])
        qc.h(q1[2])
        qc.h(q2[0])
        qc.h(q2[1])
        qc.cx(q1[0], q1[1])
        qc.cx(q1[1], q2[1])
        qc.cx(q2[1], q1[2])
        qc.cx(q1[2], q2[0])
        # Internally calls op.condition_bits
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(qc.num_connected_components(), 1)

    def test_circuit_connected_components_multi_reg2(self):
        """Test tensor factors works over multi registers #2."""

        # q1_0: ──■────────────
        #         │
        # q1_1: ──┼─────────■──
        #         │  ┌───┐  │
        # q1_2: ──┼──┤ X ├──┼──
        #         │  └─┬─┘┌─┴─┐
        # q2_0: ──┼────■──┤ X ├
        #       ┌─┴─┐     └───┘
        # q2_1: ┤ X ├──────────
        #       └───┘
        q1 = QuantumRegister(3, "q1")
        q2 = QuantumRegister(2, "q2")
        qc = QuantumCircuit(q1, q2)
        qc.cx(q1[0], q2[1])
        qc.cx(q2[0], q1[2])
        qc.cx(q1[1], q2[0])
        # Internally calls op.condition_bits
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(qc.num_connected_components(), 2)

    def test_circuit_connected_components_disconnected(self):
        """Test tensor factors works with 2q subspaces."""

        # q1_0: ──■──────────────────────
        #         │
        # q1_1: ──┼────■─────────────────
        #         │    │
        # q1_2: ──┼────┼────■────────────
        #         │    │    │
        # q1_3: ──┼────┼────┼────■───────
        #         │    │    │    │
        # q1_4: ──┼────┼────┼────┼────■──
        #         │    │    │    │  ┌─┴─┐
        # q2_0: ──┼────┼────┼────┼──┤ X ├
        #         │    │    │  ┌─┴─┐└───┘
        # q2_1: ──┼────┼────┼──┤ X ├─────
        #         │    │  ┌─┴─┐└───┘
        # q2_2: ──┼────┼──┤ X ├──────────
        #         │  ┌─┴─┐└───┘
        # q2_3: ──┼──┤ X ├───────────────
        #       ┌─┴─┐└───┘
        # q2_4: ┤ X ├────────────────────
        #       └───┘
        q1 = QuantumRegister(5, "q1")
        q2 = QuantumRegister(5, "q2")
        qc = QuantumCircuit(q1, q2)
        qc.cx(q1[0], q2[4])
        qc.cx(q1[1], q2[3])
        qc.cx(q1[2], q2[2])
        qc.cx(q1[3], q2[1])
        qc.cx(q1[4], q2[0])
        # Internally calls op.condition_bits
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(qc.num_connected_components(), 5)

    def test_circuit_connected_components_with_clbits(self):
        """Test tensor components with classical register."""

        #      ┌───┐┌─┐
        # q_0: ┤ H ├┤M├─────────
        #      ├───┤└╥┘┌─┐
        # q_1: ┤ H ├─╫─┤M├──────
        #      ├───┤ ║ └╥┘┌─┐
        # q_2: ┤ H ├─╫──╫─┤M├───
        #      ├───┤ ║  ║ └╥┘┌─┐
        # q_3: ┤ H ├─╫──╫──╫─┤M├
        #      └───┘ ║  ║  ║ └╥┘
        # c: 4/══════╩══╩══╩══╩═
        #           0  1  2  3
        size = 4
        q = QuantumRegister(size, "q")
        c = ClassicalRegister(size, "c")
        qc = QuantumCircuit(q, c)
        qc.h(q[0])
        qc.h(q[1])
        qc.h(q[2])
        qc.h(q[3])
        qc.measure(q[0], c[0])
        qc.measure(q[1], c[1])
        qc.measure(q[2], c[2])
        qc.measure(q[3], c[3])
        # Internally calls op.condition_bits
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(qc.num_connected_components(), 4)

    def test_circuit_connected_components_with_cond(self):
        """Test tensor components with one conditional gate."""

        #      ┌───┐┌─┐
        # q_0: ┤ H ├┤M├───■────────────
        #      ├───┤└╥┘   │   ┌─┐
        # q_1: ┤ H ├─╫────┼───┤M├──────
        #      ├───┤ ║    │   └╥┘┌─┐
        # q_2: ┤ H ├─╫────┼────╫─┤M├───
        #      ├───┤ ║  ┌─┴─┐  ║ └╥┘┌─┐
        # q_3: ┤ H ├─╫──┤ X ├──╫──╫─┤M├
        #      └───┘ ║  └─╥─┘  ║  ║ └╥┘
        #            ║ ┌──╨──┐ ║  ║  ║
        # c: 4/══════╩═╡ 0x2 ╞═╩══╩══╩═
        #            0 └─────┘ 1  2  3
        size = 4
        q = QuantumRegister(size, "q")
        c = ClassicalRegister(size, "c")
        qc = QuantumCircuit(q, c)
        qc.h(q[0])
        qc.h(q[1])
        qc.h(q[2])
        qc.h(q[3])
        qc.measure(q[0], c[0])
        with self.assertWarns(DeprecationWarning):
            qc.cx(q[0], q[3]).c_if(c, 2)
        qc.measure(q[1], c[1])
        qc.measure(q[2], c[2])
        qc.measure(q[3], c[3])
        # Internally calls op.condition_bits
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(qc.num_connected_components(), 1)

    def test_circuit_connected_components_with_cond2(self):
        """Test tensor components with two conditional gates."""

        #      ┌───┐ ┌───┐
        # q_0: ┤ H ├─┤ H ├────────
        #      ├───┤ └─╥─┘
        # q_1: ┤ H ├───╫──────■───
        #      ├───┤   ║    ┌─┴─┐
        # q_2: ┤ H ├───╫────┤ X ├─
        #      ├───┤   ║    └─╥─┘
        # q_3: ┤ H ├───╫──────╫───
        #      └───┘┌──╨──┐┌──╨──┐
        # c: 8/═════╡ 0x0 ╞╡ 0x4 ╞
        #           └─────┘└─────┘
        size = 4
        q = QuantumRegister(size, "q")
        c = ClassicalRegister(2 * size, "c")
        qc = QuantumCircuit(q, c)
        qc.h(q[0])
        qc.h(q[1])
        qc.h(q[2])
        qc.h(q[3])
        with self.assertWarns(DeprecationWarning):
            qc.h(0).c_if(c, 0)
        with self.assertWarns(DeprecationWarning):
            qc.cx(1, 2).c_if(c, 4)
        # Internally calls op.condition_bits
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(qc.num_connected_components(), 2)

    def test_circuit_connected_components_with_cond3(self):
        """Test tensor components with three conditional gates and measurements."""

        #       ┌───┐┌─┐ ┌───┐
        # q0_0: ┤ H ├┤M├─┤ H ├──────────────────
        #       ├───┤└╥┘ └─╥─┘
        # q0_1: ┤ H ├─╫────╫──────■─────────────
        #       ├───┤ ║    ║    ┌─┴─┐ ┌─┐
        # q0_2: ┤ H ├─╫────╫────┤ X ├─┤M├───────
        #       ├───┤ ║    ║    └─╥─┘ └╥┘ ┌───┐
        # q0_3: ┤ H ├─╫────╫──────╫────╫──┤ X ├─
        #       └───┘ ║    ║      ║    ║  └─╥─┘
        #             ║ ┌──╨──┐┌──╨──┐ ║ ┌──╨──┐
        # c0: 4/══════╩═╡ 0x0 ╞╡ 0x1 ╞═╩═╡ 0x2 ╞
        #             0 └─────┘└─────┘ 2 └─────┘
        size = 4
        q = QuantumRegister(size)
        c = ClassicalRegister(size)
        qc = QuantumCircuit(q, c)
        qc.h(q[0])
        qc.h(q[1])
        qc.h(q[2])
        qc.h(q[3])
        qc.measure(q[0], c[0])
        with self.assertWarns(DeprecationWarning):
            qc.h(q[0]).c_if(c, 0)
        with self.assertWarns(DeprecationWarning):
            qc.cx(q[1], q[2]).c_if(c, 1)
        qc.measure(q[2], c[2])
        with self.assertWarns(DeprecationWarning):
            qc.x(q[3]).c_if(c, 2)
        # Internally calls op.condition_bits
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(qc.num_connected_components(), 1)

    def test_circuit_connected_components_with_bit_cond(self):
        """Test tensor components with one single bit conditional gate."""

        #      ┌───┐┌─┐
        # q_0: ┤ H ├┤M├───────────■────────
        #      ├───┤└╥┘┌─┐        │
        # q_1: ┤ H ├─╫─┤M├────────┼────────
        #      ├───┤ ║ └╥┘┌─┐     │
        # q_2: ┤ H ├─╫──╫─┤M├─────┼────────
        #      ├───┤ ║  ║ └╥┘   ┌─┴─┐   ┌─┐
        # q_3: ┤ H ├─╫──╫──╫────┤ X ├───┤M├
        #      └───┘ ║  ║  ║    └─╥─┘   └╥┘
        #            ║  ║  ║ ┌────╨────┐ ║
        # c: 4/══════╩══╩══╩═╡ c_0=0x1 ╞═╩═
        #            0  1  2 └─────────┘ 3
        size = 4
        q = QuantumRegister(size, "q")
        c = ClassicalRegister(size, "c")
        qc = QuantumCircuit(q, c)
        qc.h(q[0])
        qc.h(q[1])
        qc.h(q[2])
        qc.h(q[3])
        qc.measure(q[0], c[0])
        with self.assertWarns(DeprecationWarning):
            qc.cx(q[0], q[3]).c_if(c[0], True)
        qc.measure(q[1], c[1])
        qc.measure(q[2], c[2])
        qc.measure(q[3], c[3])
        # Internally calls op.condition_bits
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(qc.num_connected_components(), 3)

    def test_circuit_connected_components_with_bit_cond2(self):
        """Test tensor components with two bit conditional gates."""

        #      ┌───┐   ┌───┐                 ┌───┐
        # q_0: ┤ H ├───┤ H ├─────────────────┤ X ├───
        #      ├───┤   └─╥─┘                 └─┬─┘
        # q_1: ┤ H ├─────╫─────────────────────■─────
        #      ├───┤     ║                     ║
        # q_2: ┤ H ├─────╫──────────■──────────╫─────
        #      ├───┤     ║          │          ║
        # q_3: ┤ H ├─────╫──────────■──────────╫─────
        #      └───┘┌────╨────┐┌────╨────┐┌────╨────┐
        # c: 6/═════╡ c_1=0x1 ╞╡ c_0=0x1 ╞╡ c_4=0x0 ╞
        #           └─────────┘└─────────┘└─────────┘
        size = 4
        q = QuantumRegister(size, "q")
        c = ClassicalRegister(size + 2, "c")
        qc = QuantumCircuit(q, c)
        qc.h(q[0])
        qc.h(q[1])
        qc.h(q[2])
        qc.h(q[3])
        with self.assertWarns(DeprecationWarning):
            qc.h(0).c_if(c[1], True)
        with self.assertWarns(DeprecationWarning):
            qc.cx(1, 0).c_if(c[4], False)
        with self.assertWarns(DeprecationWarning):
            qc.cz(2, 3).c_if(c[0], True)
        # Internally calls op.condition_bits
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(qc.num_connected_components(), 5)

    def test_circuit_connected_components_with_bit_cond3(self):
        """Test tensor components with register and bit conditional gates."""

        #       ┌───┐   ┌───┐
        # q0_0: ┤ H ├───┤ H ├───────────────────────
        #       ├───┤   └─╥─┘
        # q0_1: ┤ H ├─────╫─────────■───────────────
        #       ├───┤     ║       ┌─┴─┐
        # q0_2: ┤ H ├─────╫───────┤ X ├─────────────
        #       ├───┤     ║       └─╥─┘    ┌───┐
        # q0_3: ┤ H ├─────╫─────────╫──────┤ X ├────
        #       └───┘     ║         ║      └─╥─┘
        #            ┌────╨─────┐┌──╨──┐┌────╨─────┐
        # c0: 4/═════╡ c0_0=0x1 ╞╡ 0x1 ╞╡ c0_2=0x1 ╞
        #            └──────────┘└─────┘└──────────┘
        size = 4
        q = QuantumRegister(size)
        c = ClassicalRegister(size)
        qc = QuantumCircuit(q, c)
        qc.h(q[0])
        qc.h(q[1])
        qc.h(q[2])
        qc.h(q[3])
        with self.assertWarns(DeprecationWarning):
            qc.h(q[0]).c_if(c[0], True)
        with self.assertWarns(DeprecationWarning):
            qc.cx(q[1], q[2]).c_if(c, 1)
        with self.assertWarns(DeprecationWarning):
            qc.x(q[3]).c_if(c[2], True)
        # Internally calls op.condition_bits
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(qc.num_connected_components(), 1)

    def test_circuit_unitary_factors1(self):
        """Test unitary factors empty circuit."""
        size = 4
        q = QuantumRegister(size, "q")
        c = ClassicalRegister(size, "c")
        qc = QuantumCircuit(q, c)
        self.assertEqual(qc.num_unitary_factors(), 4)

    def test_circuit_unitary_factors2(self):
        """Test unitary factors multi qregs"""
        q1 = QuantumRegister(2, "q1")
        q2 = QuantumRegister(2, "q2")
        c = ClassicalRegister(4, "c")
        qc = QuantumCircuit(q1, q2, c)
        self.assertEqual(qc.num_unitary_factors(), 4)

    def test_circuit_unitary_factors3(self):
        """Test unitary factors measurements and conditionals."""

        #      ┌───┐                                      ┌─┐
        # q_0: ┤ H ├────────■──────────■────■──────────■──┤M├───
        #      ├───┤        │          │    │  ┌─┐     │  └╥┘
        # q_1: ┤ H ├──■─────┼─────■────┼────┼──┤M├─────┼───╫────
        #      ├───┤┌─┴─┐   │   ┌─┴─┐  │    │  └╥┘┌─┐  │   ║
        # q_2: ┤ H ├┤ X ├───┼───┤ X ├──┼────┼───╫─┤M├──┼───╫────
        #      ├───┤└───┘ ┌─┴─┐ └───┘┌─┴─┐┌─┴─┐ ║ └╥┘┌─┴─┐ ║ ┌─┐
        # q_3: ┤ H ├──────┤ X ├──────┤ X ├┤ X ├─╫──╫─┤ X ├─╫─┤M├
        #      └───┘      └─╥─┘      └───┘└───┘ ║  ║ └───┘ ║ └╥┘
        #                ┌──╨──┐                ║  ║       ║  ║
        # c: 4/══════════╡ 0x2 ╞════════════════╩══╩═══════╩══╩═
        #                └─────┘                1  2       0  3
        size = 4
        q = QuantumRegister(size, "q")
        c = ClassicalRegister(size, "c")
        qc = QuantumCircuit(q, c)
        qc.h(q[0])
        qc.h(q[1])
        qc.h(q[2])
        qc.h(q[3])
        qc.cx(q[1], q[2])
        qc.cx(q[1], q[2])
        with self.assertWarns(DeprecationWarning):
            qc.cx(q[0], q[3]).c_if(c, 2)
        qc.cx(q[0], q[3])
        qc.cx(q[0], q[3])
        qc.cx(q[0], q[3])
        qc.measure(q[0], c[0])
        qc.measure(q[1], c[1])
        qc.measure(q[2], c[2])
        qc.measure(q[3], c[3])
        self.assertEqual(qc.num_unitary_factors(), 2)

    def test_circuit_unitary_factors4(self):
        """Test unitary factors measurements go to same cbit."""

        #      ┌───┐┌─┐
        # q_0: ┤ H ├┤M├─────────
        #      ├───┤└╥┘┌─┐
        # q_1: ┤ H ├─╫─┤M├──────
        #      ├───┤ ║ └╥┘┌─┐
        # q_2: ┤ H ├─╫──╫─┤M├───
        #      ├───┤ ║  ║ └╥┘┌─┐
        # q_3: ┤ H ├─╫──╫──╫─┤M├
        #      └───┘ ║  ║  ║ └╥┘
        # q_4: ──────╫──╫──╫──╫─
        #            ║  ║  ║  ║
        # c: 5/══════╩══╩══╩══╩═
        #            0  0  0  0
        size = 5
        q = QuantumRegister(size, "q")
        c = ClassicalRegister(size, "c")
        qc = QuantumCircuit(q, c)
        qc.h(q[0])
        qc.h(q[1])
        qc.h(q[2])
        qc.h(q[3])
        qc.measure(q[0], c[0])
        qc.measure(q[1], c[0])
        qc.measure(q[2], c[0])
        qc.measure(q[3], c[0])
        self.assertEqual(qc.num_unitary_factors(), 5)

    def test_num_qubits_qubitless_circuit(self):
        """Check output in absence of qubits."""
        c_reg = ClassicalRegister(3)
        circ = QuantumCircuit(c_reg)
        self.assertEqual(circ.num_qubits, 0)

    def test_num_qubits_qubitfull_circuit(self):
        """Check output in presence of qubits"""
        q_reg = QuantumRegister(4)
        c_reg = ClassicalRegister(3)
        circ = QuantumCircuit(q_reg, c_reg)
        self.assertEqual(circ.num_qubits, 4)

    def test_num_qubits_registerless_circuit(self):
        """Check output for circuits with direct argument for qubits."""
        circ = QuantumCircuit(5)
        self.assertEqual(circ.num_qubits, 5)

    def test_num_qubits_multiple_register_circuit(self):
        """Check output for circuits with multiple quantum registers."""
        q_reg1 = QuantumRegister(5)
        q_reg2 = QuantumRegister(6)
        q_reg3 = QuantumRegister(7)
        circ = QuantumCircuit(q_reg1, q_reg2, q_reg3)
        self.assertEqual(circ.num_qubits, 18)

    def test_calibrations_basis_gates(self):
        """Check if the calibrations for basis gates provided are added correctly."""
        circ = QuantumCircuit(2)

        with self.assertWarns(DeprecationWarning):
            with pulse.build() as q0_x180:
                pulse.play(pulse.library.Gaussian(20, 1.0, 3.0), pulse.DriveChannel(0))
            with pulse.build() as q1_y90:
                pulse.play(pulse.library.Gaussian(20, -1.0, 3.0), pulse.DriveChannel(1))

            # Add calibration
            circ.add_calibration(RXGate(3.14), [0], q0_x180)
            circ.add_calibration(RYGate(1.57), [1], q1_y90)

            self.assertEqual(set(circ.calibrations.keys()), {"rx", "ry"})
            self.assertEqual(set(circ.calibrations["rx"].keys()), {((0,), (3.14,))})
            self.assertEqual(set(circ.calibrations["ry"].keys()), {((1,), (1.57,))})
            self.assertEqual(
                circ.calibrations["rx"][((0,), (3.14,))].instructions, q0_x180.instructions
            )
            self.assertEqual(
                circ.calibrations["ry"][((1,), (1.57,))].instructions, q1_y90.instructions
            )

    def test_calibrations_custom_gates(self):
        """Check if the calibrations for custom gates with params provided are added correctly."""
        circ = QuantumCircuit(3)

        with self.assertWarns(DeprecationWarning):
            with pulse.build() as q0_x180:
                pulse.play(pulse.library.Gaussian(20, 1.0, 3.0), pulse.DriveChannel(0))

            # Add calibrations with a custom gate 'rxt'
            circ.add_calibration("rxt", [0], q0_x180, params=[1.57, 3.14, 4.71])

            self.assertEqual(set(circ.calibrations.keys()), {"rxt"})
            self.assertEqual(set(circ.calibrations["rxt"].keys()), {((0,), (1.57, 3.14, 4.71))})
            self.assertEqual(
                circ.calibrations["rxt"][((0,), (1.57, 3.14, 4.71))].instructions,
                q0_x180.instructions,
            )

    def test_calibrations_no_params(self):
        """Check calibrations if the no params is provided with just gate name."""
        circ = QuantumCircuit(3)

        with self.assertWarns(DeprecationWarning):
            with pulse.build() as q0_x180:
                pulse.play(pulse.library.Gaussian(20, 1.0, 3.0), pulse.DriveChannel(0))

            circ.add_calibration("h", [0], q0_x180)

            self.assertEqual(set(circ.calibrations.keys()), {"h"})
            self.assertEqual(set(circ.calibrations["h"].keys()), {((0,), ())})
            self.assertEqual(circ.calibrations["h"][((0,), ())].instructions, q0_x180.instructions)

    def test_has_calibration_for(self):
        """Test that `has_calibration_for` returns a correct answer."""
        qc = QuantumCircuit(3)

        with self.assertWarns(DeprecationWarning):
            with pulse.build() as q0_x180:
                pulse.play(pulse.library.Gaussian(20, 1.0, 3.0), pulse.DriveChannel(0))
            qc.add_calibration("h", [0], q0_x180)

        qc.h(0)
        qc.h(1)

        with self.assertWarns(DeprecationWarning):
            self.assertTrue(qc.has_calibration_for(qc.data[0]))
            self.assertFalse(qc.has_calibration_for(qc.data[1]))

    def test_has_calibration_for_legacy(self):
        """Test that `has_calibration_for` returns a correct answer when presented with a legacy 3
        tuple."""
        qc = QuantumCircuit(3)

        with self.assertWarns(DeprecationWarning):
            with pulse.build() as q0_x180:
                pulse.play(pulse.library.Gaussian(20, 1.0, 3.0), pulse.DriveChannel(0))
            qc.add_calibration("h", [0], q0_x180)

        qc.h(0)
        qc.h(1)

        with self.assertWarns(DeprecationWarning):
            self.assertTrue(
                qc.has_calibration_for(
                    (qc.data[0].operation, list(qc.data[0].qubits), list(qc.data[0].clbits))
                )
            )
            self.assertFalse(
                qc.has_calibration_for(
                    (qc.data[1].operation, list(qc.data[1].qubits), list(qc.data[1].clbits))
                )
            )

    def test_metadata_copy_does_not_share_state(self):
        """Verify mutating the metadata of a circuit copy does not impact original."""
        # ref: https://github.com/Qiskit/qiskit-terra/issues/6057

        qc1 = QuantumCircuit(1)
        qc1.metadata = {"a": 0}

        qc2 = qc1.copy()
        qc2.metadata["a"] = 1000

        self.assertEqual(qc1.metadata["a"], 0)

    def test_metadata_is_dict(self):
        """Verify setting metadata to None in the constructor results in an empty dict."""
        qc = QuantumCircuit(1)
        metadata1 = qc.metadata
        self.assertEqual(metadata1, {})

    def test_metadata_raises(self):
        """Test that we must set metadata to a dict."""
        qc = QuantumCircuit(1)
        with self.assertRaises(TypeError):
            qc.metadata = 1

    def test_scheduling(self):
        """Test cannot return schedule information without scheduling."""
        qc = QuantumCircuit(2)
        qc.h(0)
        qc.cx(0, 1)

        with self.assertRaises(AttributeError):
            _ = qc.op_start_times


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2019, 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test cases for the pulse scheduler passes."""

import numpy as np
from numpy import pi
from qiskit import QuantumRegister, ClassicalRegister, QuantumCircuit, schedule
from qiskit.circuit import Gate, Parameter
from qiskit.circuit.library import U1Gate, U2Gate, U3Gate, SXGate
from qiskit.exceptions import QiskitError
from qiskit.pulse import (
    Schedule,
    DriveChannel,
    AcquireChannel,
    Acquire,
    MeasureChannel,
    MemorySlot,
    Gaussian,
    GaussianSquare,
    Play,
    Waveform,
    transforms,
)
from qiskit.pulse import build, macros, play, InstructionScheduleMap
from qiskit.providers.fake_provider import (
    FakeBackend,
    FakeOpenPulse2Q,
    FakeOpenPulse3Q,
    GenericBackendV2,
)
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class TestBasicSchedule(QiskitTestCase):
    """Scheduling tests."""

    def setUp(self):
        super().setUp()
        with self.assertWarns(DeprecationWarning):
            self.backend = FakeOpenPulse2Q()
        self.inst_map = self.backend.defaults().instruction_schedule_map

    def test_unavailable_defaults(self):
        """Test backend with unavailable defaults."""
        qr = QuantumRegister(1)
        qc = QuantumCircuit(qr)
        with self.assertWarns(DeprecationWarning):
            backend = FakeBackend(None)
        backend.defaults = backend.configuration
        with self.assertWarns(DeprecationWarning):
            self.assertRaises(QiskitError, lambda: schedule(qc, backend))

    def test_alap_pass(self):
        """Test ALAP scheduling."""

        #       ┌───────────────┐                    ░      ┌─┐
        # q0_0: ┤ U2(3.14,1.57) ├────────────────────░───■──┤M├───
        #       └┬──────────────┤ ░ ┌──────────────┐ ░ ┌─┴─┐└╥┘┌─┐
        # q0_1: ─┤ U2(0.5,0.25) ├─░─┤ U2(0.5,0.25) ├─░─┤ X ├─╫─┤M├
        #        └──────────────┘ ░ └──────────────┘ ░ └───┘ ║ └╥┘
        # c0: 2/═════════════════════════════════════════════╩══╩═
        #                                                    0  1
        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc = QuantumCircuit(q, c)
        qc.append(U2Gate(3.14, 1.57), [q[0]])
        qc.append(U2Gate(0.5, 0.25), [q[1]])
        qc.barrier(q[1])
        qc.append(U2Gate(0.5, 0.25), [q[1]])
        qc.barrier(q[0], [q[1]])
        qc.cx(q[0], q[1])
        qc.measure(q, c)
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend)
        # X pulse on q0 should end at the start of the CNOT
        with self.assertWarns(DeprecationWarning):
            expected = Schedule(
                (2, self.inst_map.get("u2", [0], 3.14, 1.57)),
                self.inst_map.get("u2", [1], 0.5, 0.25),
                (2, self.inst_map.get("u2", [1], 0.5, 0.25)),
                (4, self.inst_map.get("cx", [0, 1])),
                (26, self.inst_map.get("measure", [0, 1])),
            )
        for actual, expected in zip(sched.instructions, expected.instructions):
            self.assertEqual(actual[0], expected[0])
            self.assertEqual(actual[1], expected[1])

    def test_single_circuit_list_schedule(self):
        """Test that passing a single circuit list to schedule() returns a list."""
        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc = QuantumCircuit(q, c)
        with self.assertWarns(DeprecationWarning):
            sched = schedule([qc], self.backend, method="alap")
            expected = Schedule()
        self.assertIsInstance(sched, list)
        self.assertEqual(sched[0].instructions, expected.instructions)

    def test_alap_with_barriers(self):
        """Test that ALAP respects barriers on new qubits."""
        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc = QuantumCircuit(q, c)
        qc.append(U2Gate(0, 0), [q[0]])
        qc.barrier(q[0], q[1])
        qc.append(U2Gate(0, 0), [q[1]])
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="alap")
            expected = Schedule(
                self.inst_map.get("u2", [0], 0, 0), (2, self.inst_map.get("u2", [1], 0, 0))
            )
        for actual, expected in zip(sched.instructions, expected.instructions):
            self.assertEqual(actual[0], expected[0])
            self.assertEqual(actual[1], expected[1])

    def test_empty_circuit_schedule(self):
        """Test empty circuit being scheduled."""
        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc = QuantumCircuit(q, c)
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="alap")
            expected = Schedule()
        self.assertEqual(sched.instructions, expected.instructions)

    def test_alap_aligns_end(self):
        """Test that ALAP always acts as though there is a final global barrier."""
        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc = QuantumCircuit(q, c)
        qc.append(U3Gate(0, 0, 0), [q[0]])
        qc.append(U2Gate(0, 0), [q[1]])
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="alap")
            expected_sched = Schedule(
                (2, self.inst_map.get("u2", [1], 0, 0)), self.inst_map.get("u3", [0], 0, 0, 0)
            )
        for actual, expected in zip(sched.instructions, expected_sched.instructions):
            self.assertEqual(actual[0], expected[0])
            self.assertEqual(actual[1], expected[1])
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(
                sched.ch_duration(DriveChannel(0)), expected_sched.ch_duration(DriveChannel(1))
            )

    def test_asap_pass(self):
        """Test ASAP scheduling."""

        #       ┌───────────────┐                    ░      ┌─┐
        # q0_0: ┤ U2(3.14,1.57) ├────────────────────░───■──┤M├───
        #       └┬──────────────┤ ░ ┌──────────────┐ ░ ┌─┴─┐└╥┘┌─┐
        # q0_1: ─┤ U2(0.5,0.25) ├─░─┤ U2(0.5,0.25) ├─░─┤ X ├─╫─┤M├
        #        └──────────────┘ ░ └──────────────┘ ░ └───┘ ║ └╥┘
        # c0: 2/═════════════════════════════════════════════╩══╩═
        #                                                    0  1
        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc = QuantumCircuit(q, c)
        qc.append(U2Gate(3.14, 1.57), [q[0]])
        qc.append(U2Gate(0.5, 0.25), [q[1]])
        qc.barrier(q[1])
        qc.append(U2Gate(0.5, 0.25), [q[1]])
        qc.barrier(q[0], q[1])
        qc.cx(q[0], q[1])
        qc.measure(q, c)
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="as_soon_as_possible")
            # X pulse on q0 should start at t=0
            expected = Schedule(
                self.inst_map.get("u2", [0], 3.14, 1.57),
                self.inst_map.get("u2", [1], 0.5, 0.25),
                (2, self.inst_map.get("u2", [1], 0.5, 0.25)),
                (4, self.inst_map.get("cx", [0, 1])),
                (26, self.inst_map.get("measure", [0, 1])),
            )
        for actual, expected in zip(sched.instructions, expected.instructions):
            self.assertEqual(actual[0], expected[0])
            self.assertEqual(actual[1], expected[1])

    def test_alap_resource_respecting(self):
        """Test that the ALAP pass properly respects busy resources when backwards scheduling.
        For instance, a CX on 0 and 1 followed by an X on only 1 must respect both qubits'
        timeline."""
        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc = QuantumCircuit(q, c)
        qc.cx(q[0], q[1])
        qc.append(U2Gate(0.5, 0.25), [q[1]])
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="as_late_as_possible")
        insts = sched.instructions
        self.assertEqual(insts[0][0], 0)
        self.assertEqual(insts[6][0], 22)

        qc = QuantumCircuit(q, c)
        qc.cx(q[0], q[1])
        qc.append(U2Gate(0.5, 0.25), [q[1]])
        qc.measure(q, c)
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="as_late_as_possible")
        self.assertEqual(sched.instructions[-1][0], 24)

    def test_inst_map_schedules_unaltered(self):
        """Test that forward scheduling doesn't change relative timing with a command."""
        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc = QuantumCircuit(q, c)
        qc.cx(q[0], q[1])
        with self.assertWarns(DeprecationWarning):
            sched1 = schedule(qc, self.backend, method="as_soon_as_possible")
            sched2 = schedule(qc, self.backend, method="as_late_as_possible")
        for asap, alap in zip(sched1.instructions, sched2.instructions):
            self.assertEqual(asap[0], alap[0])
            self.assertEqual(asap[1], alap[1])
        insts = sched1.instructions
        self.assertEqual(insts[0][0], 0)  # shift phase
        self.assertEqual(insts[1][0], 0)  # ym_d0
        self.assertEqual(insts[2][0], 0)  # x90p_d1
        self.assertEqual(insts[3][0], 2)  # cr90p_u0
        self.assertEqual(insts[4][0], 11)  # xp_d0
        self.assertEqual(insts[5][0], 13)  # cr90m_u0

    def test_measure_combined(self):
        """
        Test to check for measure on the same qubit which generated another measure schedule.

        The measures on different qubits are combined, but measures on the same qubit
        adds another measure to the schedule.
        """
        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc = QuantumCircuit(q, c)
        qc.append(U2Gate(3.14, 1.57), [q[0]])
        qc.cx(q[0], q[1])
        qc.measure(q[0], c[0])
        qc.measure(q[1], c[1])
        qc.measure(q[1], c[1])
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="as_soon_as_possible")
            expected = Schedule(
                self.inst_map.get("u2", [0], 3.14, 1.57),
                (2, self.inst_map.get("cx", [0, 1])),
                (24, self.inst_map.get("measure", [0, 1])),
                (34, self.inst_map.get("measure", [0, 1]).filter(channels=[MeasureChannel(1)])),
                (34, Acquire(10, AcquireChannel(1), MemorySlot(1))),
            )
        self.assertEqual(sched.instructions, expected.instructions)

    #
    def test_3q_schedule(self):
        """Test a schedule that was recommended by David McKay :D"""

        #                          ┌─────────────────┐
        # q0_0: ─────────■─────────┤ U3(3.14,1.57,0) ├────────────────────────
        #              ┌─┴─┐       └┬───────────────┬┘
        # q0_1: ───────┤ X ├────────┤ U2(3.14,1.57) ├───■─────────────────────
        #       ┌──────┴───┴──────┐ └───────────────┘ ┌─┴─┐┌─────────────────┐
        # q0_2: ┤ U2(0.778,0.122) ├───────────────────┤ X ├┤ U2(0.778,0.122) ├
        #       └─────────────────┘                   └───┘└─────────────────┘
        with self.assertWarns(DeprecationWarning):
            backend = FakeOpenPulse3Q()
        inst_map = backend.defaults().instruction_schedule_map
        q = QuantumRegister(3)
        c = ClassicalRegister(3)
        qc = QuantumCircuit(q, c)
        qc.cx(q[0], q[1])
        qc.append(U2Gate(0.778, 0.122), [q[2]])
        qc.append(U3Gate(3.14, 1.57, 0), [q[0]])
        qc.append(U2Gate(3.14, 1.57), [q[1]])
        qc.cx(q[1], q[2])
        qc.append(U2Gate(0.778, 0.122), [q[2]])
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, backend)
            expected = Schedule(
                inst_map.get("cx", [0, 1]),
                (22, inst_map.get("u2", [1], 3.14, 1.57)),
                (22, inst_map.get("u2", [2], 0.778, 0.122)),
                (24, inst_map.get("cx", [1, 2])),
                (44, inst_map.get("u3", [0], 3.14, 1.57, 0)),
                (46, inst_map.get("u2", [2], 0.778, 0.122)),
            )
        for actual, expected in zip(sched.instructions, expected.instructions):
            self.assertEqual(actual[0], expected[0])
            self.assertEqual(actual[1], expected[1])

    def test_schedule_multi(self):
        """Test scheduling multiple circuits at once."""
        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc0 = QuantumCircuit(q, c)
        qc0.cx(q[0], q[1])
        qc1 = QuantumCircuit(q, c)
        qc1.cx(q[0], q[1])
        with self.assertWarns(DeprecationWarning):
            schedules = schedule([qc0, qc1], self.backend)
            expected_insts = schedule(qc0, self.backend).instructions
        for actual, expected in zip(schedules[0].instructions, expected_insts):
            self.assertEqual(actual[0], expected[0])
            self.assertEqual(actual[1], expected[1])

    def test_circuit_name_kept(self):
        """Test that the new schedule gets its name from the circuit."""
        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc = QuantumCircuit(q, c, name="CIRCNAME")
        qc.cx(q[0], q[1])
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="asap")
        self.assertEqual(sched.name, qc.name)
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="alap")
        self.assertEqual(sched.name, qc.name)

    def test_can_add_gates_into_free_space(self):
        """The scheduler does some time bookkeeping to know when qubits are free to be
        scheduled. Make sure this works for qubits that are used in the future. This was
        a bug, uncovered by this example:

           q0 =  - - - - |X|
           q1 = |X| |u2| |X|

        In ALAP scheduling, the next operation on qubit 0 would be added at t=0 rather
        than immediately before the X gate.
        """
        qr = QuantumRegister(2)
        qc = QuantumCircuit(qr)
        for i in range(2):
            qc.append(U2Gate(0, 0), [qr[i]])
            qc.append(U1Gate(3.14), [qr[i]])
            qc.append(U2Gate(0, 0), [qr[i]])
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="alap")
            expected = Schedule(
                self.inst_map.get("u2", [0], 0, 0),
                self.inst_map.get("u2", [1], 0, 0),
                (2, self.inst_map.get("u1", [0], 3.14)),
                (2, self.inst_map.get("u1", [1], 3.14)),
                (2, self.inst_map.get("u2", [0], 0, 0)),
                (2, self.inst_map.get("u2", [1], 0, 0)),
            )
        for actual, expected in zip(sched.instructions, expected.instructions):
            self.assertEqual(actual[0], expected[0])
            self.assertEqual(actual[1], expected[1])

    def test_barriers_in_middle(self):
        """As a follow on to `test_can_add_gates_into_free_space`, similar issues
        arose for barriers, specifically.
        """
        qr = QuantumRegister(2)
        qc = QuantumCircuit(qr)
        for i in range(2):
            qc.append(U2Gate(0, 0), [qr[i]])
            qc.barrier(qr[i])
            qc.append(U1Gate(3.14), [qr[i]])
            qc.barrier(qr[i])
            qc.append(U2Gate(0, 0), [qr[i]])
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="alap")
            expected = Schedule(
                self.inst_map.get("u2", [0], 0, 0),
                self.inst_map.get("u2", [1], 0, 0),
                (2, self.inst_map.get("u1", [0], 3.14)),
                (2, self.inst_map.get("u1", [1], 3.14)),
                (2, self.inst_map.get("u2", [0], 0, 0)),
                (2, self.inst_map.get("u2", [1], 0, 0)),
            )
        for actual, expected in zip(sched.instructions, expected.instructions):
            self.assertEqual(actual[0], expected[0])
            self.assertEqual(actual[1], expected[1])

    def test_parametric_input(self):
        """Test that scheduling works with parametric pulses as input."""
        qr = QuantumRegister(1)
        qc = QuantumCircuit(qr)
        qc.append(Gate("gauss", 1, []), qargs=[qr[0]])
        with self.assertWarns(DeprecationWarning):
            custom_gauss = Schedule(
                Play(Gaussian(duration=25, sigma=4, amp=0.5, angle=pi / 2), DriveChannel(0))
            )
        self.inst_map.add("gauss", [0], custom_gauss)
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, inst_map=self.inst_map)
        self.assertEqual(sched.instructions[0], custom_gauss.instructions[0])

    def test_pulse_gates(self):
        """Test scheduling calibrated pulse gates."""
        q = QuantumRegister(2)
        qc = QuantumCircuit(q)
        qc.append(U2Gate(0, 0), [q[0]])
        qc.barrier(q[0], q[1])
        qc.append(U2Gate(0, 0), [q[1]])
        with self.assertWarns(DeprecationWarning):
            qc.add_calibration(
                "u2", [0], Schedule(Play(Gaussian(28, 0.2, 4), DriveChannel(0))), [0, 0]
            )
            qc.add_calibration(
                "u2", [1], Schedule(Play(Gaussian(28, 0.2, 4), DriveChannel(1))), [0, 0]
            )

        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend)
            expected = Schedule(
                Play(Gaussian(28, 0.2, 4), DriveChannel(0)),
                (28, Schedule(Play(Gaussian(28, 0.2, 4), DriveChannel(1)))),
            )
        self.assertEqual(sched.instructions, expected.instructions)

    def test_calibrated_measurements(self):
        """Test scheduling calibrated measurements."""
        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc = QuantumCircuit(q, c)
        qc.append(U2Gate(0, 0), [q[0]])
        qc.measure(q[0], c[0])

        with self.assertWarns(DeprecationWarning):
            meas_sched = Play(Gaussian(1200, 0.2, 4), MeasureChannel(0))
            meas_sched |= Acquire(1200, AcquireChannel(0), MemorySlot(0))
            qc.add_calibration("measure", [0], meas_sched)

            sched = schedule(qc, self.backend)
            expected = Schedule(self.inst_map.get("u2", [0], 0, 0), (2, meas_sched))
        self.assertEqual(sched.instructions, expected.instructions)

    def test_subset_calibrated_measurements(self):
        """Test that measurement calibrations can be added and used for some qubits, even
        if the other qubits do not also have calibrated measurements."""
        qc = QuantumCircuit(3, 3)
        qc.measure(0, 0)
        qc.measure(1, 1)
        qc.measure(2, 2)
        meas_scheds = []
        for qubit in [0, 2]:
            with self.assertWarns(DeprecationWarning):
                meas = Play(Gaussian(1200, 0.2, 4), MeasureChannel(qubit)) + Acquire(
                    1200, AcquireChannel(qubit), MemorySlot(qubit)
                )
            meas_scheds.append(meas)
            with self.assertWarns(DeprecationWarning):
                qc.add_calibration("measure", [qubit], meas)

        with self.assertWarns(DeprecationWarning):
            backend = FakeOpenPulse3Q()
            meas = macros.measure([1], backend)
            meas = meas.exclude(channels=[AcquireChannel(0), AcquireChannel(2)])
        with self.assertWarns(DeprecationWarning):
            backend = FakeOpenPulse3Q()
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, backend)
            expected = Schedule(meas_scheds[0], meas_scheds[1], meas)
        self.assertEqual(sched.instructions, expected.instructions)

    def test_clbits_of_calibrated_measurements(self):
        """Test that calibrated measurements are only used when the classical bits also match."""
        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc = QuantumCircuit(q, c)
        qc.measure(q[0], c[1])

        with self.assertWarns(DeprecationWarning):
            meas_sched = Play(Gaussian(1200, 0.2, 4), MeasureChannel(0))
            meas_sched |= Acquire(1200, AcquireChannel(0), MemorySlot(0))
            qc.add_calibration("measure", [0], meas_sched)

        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend)
        # Doesn't use the calibrated schedule because the classical memory slots do not match
        with self.assertWarns(DeprecationWarning):
            expected = Schedule(macros.measure([0], self.backend, qubit_mem_slots={0: 1}))
        self.assertEqual(sched.instructions, expected.instructions)

    def test_metadata_is_preserved_alap(self):
        """Test that circuit metadata is preserved in output schedule with alap."""
        q = QuantumRegister(2)
        qc = QuantumCircuit(q)
        qc.append(U2Gate(0, 0), [q[0]])
        qc.barrier(q[0], q[1])
        qc.append(U2Gate(0, 0), [q[1]])
        qc.metadata = {"experiment_type": "gst", "execution_number": "1234"}
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="alap")
        self.assertEqual({"experiment_type": "gst", "execution_number": "1234"}, sched.metadata)

    def test_metadata_is_preserved_asap(self):
        """Test that circuit metadata is preserved in output schedule with asap."""
        q = QuantumRegister(2)
        qc = QuantumCircuit(q)
        qc.append(U2Gate(0, 0), [q[0]])
        qc.barrier(q[0], q[1])
        qc.append(U2Gate(0, 0), [q[1]])
        qc.metadata = {"experiment_type": "gst", "execution_number": "1234"}
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="asap")
        self.assertEqual({"experiment_type": "gst", "execution_number": "1234"}, sched.metadata)

    def test_scheduler_with_params_bound(self):
        """Test scheduler with parameters defined and bound"""
        x = Parameter("x")
        qc = QuantumCircuit(2)
        qc.append(Gate("pulse_gate", 1, [x]), [0])
        with self.assertWarns(DeprecationWarning):
            expected_schedule = Schedule()
            qc.add_calibration(
                gate="pulse_gate", qubits=[0], schedule=expected_schedule, params=[x]
            )
        qc = qc.assign_parameters({x: 1})
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend)
        self.assertEqual(sched, expected_schedule)

    def test_scheduler_with_params_not_bound(self):
        """Test scheduler with parameters defined but not bound"""
        x = Parameter("amp")
        qc = QuantumCircuit(2)
        qc.append(Gate("pulse_gate", 1, [x]), [0])
        with self.assertWarns(DeprecationWarning):
            with build() as expected_schedule:
                play(Gaussian(duration=160, amp=x, sigma=40), DriveChannel(0))
            qc.add_calibration(
                gate="pulse_gate", qubits=[0], schedule=expected_schedule, params=[x]
            )
            sched = schedule(qc, self.backend)
            self.assertEqual(sched, transforms.target_qobj_transform(expected_schedule))

    def test_schedule_block_in_instmap(self):
        """Test schedule block in instmap can be scheduled."""
        duration = Parameter("duration")

        with self.assertWarns(DeprecationWarning):
            with build() as pulse_prog:
                play(Gaussian(duration, 0.1, 10), DriveChannel(0))

            instmap = InstructionScheduleMap()
        instmap.add("block_gate", (0,), pulse_prog, ["duration"])

        qc = QuantumCircuit(1)
        qc.append(Gate("block_gate", 1, [duration]), [0])
        qc.assign_parameters({duration: 100}, inplace=True)

        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, inst_map=instmap)

        with self.assertWarns(DeprecationWarning):
            ref_sched = Schedule()
            ref_sched += Play(Gaussian(100, 0.1, 10), DriveChannel(0))

        self.assertEqual(sched, ref_sched)


class TestBasicScheduleV2(QiskitTestCase):
    """Scheduling tests."""

    def setUp(self):
        super().setUp()
        with self.assertWarns(DeprecationWarning):
            self.backend = GenericBackendV2(num_qubits=3, calibrate_instructions=True, seed=42)
            self.inst_map = self.backend.instruction_schedule_map
        # self.pulse_2_samples is the pulse sequence used to calibrate "measure" in
        # GenericBackendV2. See class construction for more details.
        self.pulse_2_samples = np.linspace(0, 1.0, 32, dtype=np.complex128)

    def test_alap_pass(self):
        """Test ALAP scheduling."""

        #       ┌────┐          ░      ┌─┐
        # q0_0: ┤ √X ├──────────░───■──┤M├───
        #       ├────┤ ░ ┌────┐ ░ ┌─┴─┐└╥┘┌─┐
        # q0_1: ┤ √X ├─░─┤ √X ├─░─┤ X ├─╫─┤M├
        #       └────┘ ░ └────┘ ░ └───┘ ║ └╥┘
        # c0: 2/════════════════════════╩══╩═
        #                               0  1

        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc = QuantumCircuit(q, c)

        qc.sx(q[0])
        qc.sx(q[1])
        qc.barrier(q[1])

        qc.sx(q[1])
        qc.barrier(q[0], q[1])

        qc.cx(q[0], q[1])
        qc.measure(q, c)

        with self.assertWarns(DeprecationWarning):
            sched = schedule(circuits=qc, backend=self.backend, method="alap")

        # Since, the method of scheduling chosen here is 'as_late_as_possible'
        # so all the π/2 pulse here should be right shifted.
        #
        # Calculations:
        #   Duration of the π/2 pulse for GenericBackendV2 backend is 16dt
        #   first π/2 pulse on q0 should start at 16dt because of 'as_late_as_possible'.
        #   first π/2 pulse on q1 should start 0dt.
        #   second π/2 pulse on q1 should start with a delay of 16dt.
        #   cx pulse( pulse on drive channel, control channel) should start with a delay
        #   of 16dt+16dt.
        #   measure pulse should start with a delay of 16dt+16dt+64dt(64dt for cx gate).
        with self.assertWarns(DeprecationWarning):
            expected = Schedule(
                (0, self.inst_map.get("sx", [1])),
                (0 + 16, self.inst_map.get("sx", [0])),  # Right shifted because of alap.
                (0 + 16, self.inst_map.get("sx", [1])),
                (0 + 16 + 16, self.inst_map.get("cx", [0, 1])),
                (
                    0 + 16 + 16 + 64,
                    Play(
                        Waveform(samples=self.pulse_2_samples, name="pulse_2"),
                        MeasureChannel(0),
                        name="pulse_2",
                    ),
                ),
                (
                    0 + 16 + 16 + 64,
                    Play(
                        Waveform(samples=self.pulse_2_samples, name="pulse_2"),
                        MeasureChannel(1),
                        name="pulse_2",
                    ),
                ),
                (0 + 16 + 16 + 64, Acquire(1792, AcquireChannel(0), MemorySlot(0))),
                (0 + 16 + 16 + 64, Acquire(1792, AcquireChannel(1), MemorySlot(1))),
            )
        for actual, expected in zip(sched.instructions, expected.instructions):
            self.assertEqual(actual[0], expected[0])
            self.assertEqual(actual[1], expected[1])

    def test_single_circuit_list_schedule(self):
        """Test that passing a single circuit list to schedule() returns a list."""
        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc = QuantumCircuit(q, c)
        with self.assertWarns(DeprecationWarning):
            sched = schedule([qc], self.backend, method="alap")
            expected = Schedule()
        self.assertIsInstance(sched, list)
        self.assertEqual(sched[0].instructions, expected.instructions)

    def test_alap_with_barriers(self):
        """Test that ALAP respects barriers on new qubits."""

        #       ┌────┐ ░
        # q0_0: ┤ √X ├─░───────
        #       └────┘ ░ ┌────┐
        # q0_1: ───────░─┤ √X ├
        #              ░ └────┘
        # c0: 2/═══════════════
        #

        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc = QuantumCircuit(q, c)
        qc.append(SXGate(), [q[0]])
        qc.barrier(q[0], q[1])
        qc.append(SXGate(), [q[1]])
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="alap")
            # If there wasn't a barrier the π/2 pulse on q1 would have started from 0dt, but since,
            # there is a barrier so the π/2 pulse on q1 should start with a delay of 160dt.
            expected = Schedule(
                (0, self.inst_map.get("sx", [0])), (16, self.inst_map.get("sx", [1]))
            )
        for actual, expected in zip(sched.instructions, expected.instructions):
            self.assertEqual(actual, expected)

    def test_empty_circuit_schedule(self):
        """Test empty circuit being scheduled."""
        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc = QuantumCircuit(q, c)
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="alap")
            expected = Schedule()
        self.assertEqual(sched.instructions, expected.instructions)

    def test_alap_aligns_end(self):
        """Test that ALAP always acts as though there is a final global barrier."""
        #       ┌────┐
        # q1_0: ┤ √X ├
        #       ├────┤
        # q1_1: ┤ √X ├
        #       └────┘
        # c1: 2/══════

        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc = QuantumCircuit(q, c)
        qc.sx(q[0])
        qc.sx(q[1])
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="alap")
            expected_sched = Schedule(
                (0, self.inst_map.get("sx", [1])), (0, self.inst_map.get("sx", [0]))
            )
        for actual, expected in zip(sched.instructions, expected_sched.instructions):
            self.assertEqual(actual[0], expected[0])
            self.assertEqual(actual[1], expected[1])
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(
                sched.ch_duration(DriveChannel(0)), expected_sched.ch_duration(DriveChannel(1))
            )

    def test_asap_pass(self):
        """Test ASAP scheduling."""

        #       ┌────┐          ░      ┌─┐
        # q0_0: ┤ √X ├──────────░───■──┤M├───
        #       ├────┤ ░ ┌────┐ ░ ┌─┴─┐└╥┘┌─┐
        # q0_1: ┤ √X ├─░─┤ √X ├─░─┤ X ├─╫─┤M├
        #       └────┘ ░ └────┘ ░ └───┘ ║ └╥┘
        # c0: 2/════════════════════════╩══╩═
        #                               0  1

        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc = QuantumCircuit(q, c)

        qc.sx(q[0])
        qc.sx(q[1])
        qc.barrier(q[1])

        qc.sx(q[1])
        qc.barrier(q[0], q[1])

        qc.cx(q[0], q[1])
        qc.measure(q, c)

        with self.assertWarns(DeprecationWarning):
            sched = schedule(circuits=qc, backend=self.backend, method="asap")
        # Since, the method of scheduling chosen here is 'as_soon_as_possible'
        # so all the π/2 pulse here should be left shifted.
        #
        # Calculations:
        #   Duration of the π/2 pulse for FakePerth backend is 16dt
        #   first π/2 pulse on q0 should start at 0dt because of 'as_soon_as_possible'.
        #   first π/2 pulse on q1 should start 0dt.
        #   second π/2 pulse on q1 should start with a delay of 16dt.
        #   cx pulse( pulse on drive channel, control channel) should start with a delay
        #   of 16dt+16dt.
        #   measure pulse should start with a delay of 16dt+16dt+64dt(64dt for cx gate).
        with self.assertWarns(DeprecationWarning):
            expected = Schedule(
                (0, self.inst_map.get("sx", [1])),
                (0, self.inst_map.get("sx", [0])),  # Left shifted because of asap.
                (0 + 16, self.inst_map.get("sx", [1])),
                (0 + 16 + 16, self.inst_map.get("cx", [0, 1])),
                (
                    0 + 16 + 16 + 64,
                    Play(
                        Waveform(samples=self.pulse_2_samples, name="pulse_2"),
                        MeasureChannel(0),
                        name="pulse_2",
                    ),
                ),
                (
                    0 + 16 + 16 + 64,
                    Play(
                        Waveform(samples=self.pulse_2_samples, name="pulse_2"),
                        MeasureChannel(1),
                        name="pulse_2",
                    ),
                ),
                (0 + 16 + 16 + 64, Acquire(1792, AcquireChannel(0), MemorySlot(0))),
                (0 + 16 + 16 + 64, Acquire(1792, AcquireChannel(1), MemorySlot(1))),
            )
        for actual, expected in zip(sched.instructions, expected.instructions):
            self.assertEqual(actual[0], expected[0])
            self.assertEqual(actual[1], expected[1])

    def test_alap_resource_respecting(self):
        """Test that the ALAP pass properly respects busy resources when backwards scheduling.
        For instance, a CX on 0 and 1 followed by an X on only 1 must respect both qubits'
        timeline."""
        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc = QuantumCircuit(q, c)
        qc.cx(q[0], q[1])
        qc.sx(q[1])
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="as_late_as_possible")
        insts = sched.instructions

        # This is ShiftPhase for the cx operation.
        self.assertEqual(insts[0][0], 0)

        # It takes 4 pulse operations on DriveChannel and ControlChannel to do a
        # cx operation on this backend.
        # 64dt is duration of cx operation on this backend.
        self.assertEqual(insts[4][0], 64)

        qc = QuantumCircuit(q, c)
        qc.cx(q[0], q[1])
        qc.sx(q[1])
        qc.measure(q, c)
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="as_late_as_possible")

        # 64dt for cx operation + 16dt for sx operation
        # So, the pulses in MeasureChannel0 and 1 starts from 80dt.
        self.assertEqual(sched.instructions[-1][0], 80)

    def test_inst_map_schedules_unaltered(self):
        """Test that forward scheduling doesn't change relative timing with a command."""
        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc = QuantumCircuit(q, c)
        qc.cx(q[0], q[1])
        with self.assertWarns(DeprecationWarning):
            sched1 = schedule(qc, self.backend, method="as_soon_as_possible")
            sched2 = schedule(qc, self.backend, method="as_late_as_possible")
        for asap, alap in zip(sched1.instructions, sched2.instructions):
            self.assertEqual(asap[0], alap[0])
            self.assertEqual(asap[1], alap[1])
        insts = sched1.instructions
        self.assertEqual(insts[0][0], 0)  # ShiftPhase at DriveChannel(0) no dt required.
        self.assertEqual(insts[1][0], 0)  # ShiftPhase at ControlChannel(1) no dt required.
        self.assertEqual(insts[2][0], 0)  # Pulse pulse_2 of duration 32dt.
        self.assertEqual(insts[3][0], 0)  # Pulse pulse_3 of duration 160dt.

    def test_measure_combined(self):
        """
        Test to check for measure on the same qubit which generated another measure schedule.

        The measures on different qubits are combined, but measures on the same qubit
        adds another measure to the schedule.
        """
        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc = QuantumCircuit(q, c)
        qc.sx(q[0])
        qc.cx(q[0], q[1])
        qc.measure(q[0], c[0])
        qc.measure(q[1], c[1])
        qc.measure(q[1], c[1])

        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="as_soon_as_possible")

            expected_sched = Schedule(
                # This is the schedule to implement sx gate.
                (0, self.inst_map.get("sx", [0])),
                # This is the schedule to implement cx gate
                (0 + 16, self.inst_map.get("cx", [0, 1])),
                # This is the schedule for the measurements on qubits 0 and 1 (combined)
                (
                    0 + 16 + 64,
                    self.inst_map.get("measure", [0]).filter(
                        channels=[MeasureChannel(0), MeasureChannel(1)]
                    ),
                ),
                (
                    0 + 16 + 64,
                    self.inst_map.get("measure", [0]).filter(
                        channels=[AcquireChannel(0), AcquireChannel(1)]
                    ),
                ),
                # This is the schedule for the second measurement on qubit 1
                (
                    0 + 16 + 64 + 1792,
                    self.inst_map.get("measure", [1]).filter(channels=[MeasureChannel(1)]),
                ),
                (
                    0 + 16 + 64 + 1792,
                    self.inst_map.get("measure", [1]).filter(channels=[AcquireChannel(1)]),
                ),
            )
        self.assertEqual(sched.instructions, expected_sched.instructions)

    def test_3q_schedule(self):
        """Test a schedule that was recommended by David McKay :D"""

        #             ┌────┐
        # q0_0: ──■───┤ √X ├───────────
        #       ┌─┴─┐ ├───┬┘
        # q0_1: ┤ X ├─┤ X ├───■────────
        #       ├───┴┐└───┘ ┌─┴─┐┌────┐
        # q0_2: ┤ √X ├──────┤ X ├┤ √X ├
        #       └────┘      └───┘└────┘
        # c0: 3/═══════════════════════

        q = QuantumRegister(3)
        c = ClassicalRegister(3)
        qc = QuantumCircuit(q, c)
        qc.cx(q[0], q[1])
        qc.sx(q[0])
        qc.x(q[1])
        qc.sx(q[2])
        qc.cx(q[1], q[2])
        qc.sx(q[2])

        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="asap")
            expected = Schedule(
                (0, self.inst_map.get("cx", [0, 1])),
                (0, self.inst_map.get("sx", [2])),
                (0 + 64, self.inst_map.get("sx", [0])),
                (0 + 64, self.inst_map.get("x", [1])),
                (0 + 64 + 16, self.inst_map.get("cx", [1, 2])),
                (0 + 64 + 16 + 64, self.inst_map.get("sx", [2])),
            )
        for actual, expected in zip(sched.instructions, expected.instructions):
            self.assertEqual(actual[0], expected[0])
            self.assertEqual(actual[1], expected[1])

    def test_schedule_multi(self):
        """Test scheduling multiple circuits at once."""
        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc0 = QuantumCircuit(q, c)
        qc0.cx(q[0], q[1])
        qc1 = QuantumCircuit(q, c)
        qc1.cx(q[0], q[1])
        with self.assertWarns(DeprecationWarning):
            schedules = schedule([qc0, qc1], self.backend)
            expected_insts = schedule(qc0, self.backend).instructions
        for actual, expected in zip(schedules[0].instructions, expected_insts):
            self.assertEqual(actual[0], expected[0])
            self.assertEqual(actual[1], expected[1])

    def test_circuit_name_kept(self):
        """Test that the new schedule gets its name from the circuit."""
        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc = QuantumCircuit(q, c, name="CIRCNAME")
        qc.cx(q[0], q[1])
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="asap")
        self.assertEqual(sched.name, qc.name)
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="alap")
        self.assertEqual(sched.name, qc.name)

    def test_can_add_gates_into_free_space(self):
        """The scheduler does some time bookkeeping to know when qubits are free to be
        scheduled. Make sure this works for qubits that are used in the future. This was
        a bug, uncovered by this example:

           q0 =  - - - - |X|
           q1 = |X| |u2| |X|

        In ALAP scheduling, the next operation on qubit 0 would be added at t=0 rather
        than immediately before the X gate.
        """
        qr = QuantumRegister(2)
        qc = QuantumCircuit(qr)
        for i in range(2):
            qc.sx(qr[i])
            qc.x(qr[i])
            qc.sx(qr[i])
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="alap")
            expected = Schedule(
                (0, self.inst_map.get("sx", [0])),
                (0, self.inst_map.get("sx", [1])),
                (0 + 16, self.inst_map.get("x", [0])),
                (0 + 16, self.inst_map.get("x", [1])),
                (0 + 16 + 16, self.inst_map.get("sx", [0])),
                (0 + 16 + 16, self.inst_map.get("sx", [1])),
            )
        for actual, expected in zip(sched.instructions, expected.instructions):
            self.assertEqual(actual[0], expected[0])
            self.assertEqual(actual[1], expected[1])

    def test_barriers_in_middle(self):
        """As a follow on to `test_can_add_gates_into_free_space`, similar issues
        arose for barriers, specifically.
        """
        qr = QuantumRegister(2)
        qc = QuantumCircuit(qr)
        for i in range(2):
            qc.sx(qr[i])
            qc.barrier(qr[i])
            qc.x(qr[i])
            qc.barrier(qr[i])
            qc.sx(qr[i])
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="alap")
            expected = Schedule(
                (0, self.inst_map.get("sx", [0])),
                (0, self.inst_map.get("sx", [1])),
                (0 + 16, self.inst_map.get("x", [0])),
                (0 + 16, self.inst_map.get("x", [1])),
                (0 + 16 + 16, self.inst_map.get("sx", [0])),
                (0 + 16 + 16, self.inst_map.get("sx", [1])),
            )
        for actual, expected in zip(sched.instructions, expected.instructions):
            self.assertEqual(actual[0], expected[0])
            self.assertEqual(actual[1], expected[1])

    def test_parametric_input(self):
        """Test that scheduling works with parametric pulses as input."""
        qr = QuantumRegister(1)
        qc = QuantumCircuit(qr)
        qc.append(Gate("gauss", 1, []), qargs=[qr[0]])
        with self.assertWarns(DeprecationWarning):
            custom_gauss = Schedule(
                Play(Gaussian(duration=25, sigma=4, amp=0.5, angle=pi / 2), DriveChannel(0))
            )
        self.inst_map.add("gauss", [0], custom_gauss)
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, inst_map=self.inst_map)
        self.assertEqual(sched.instructions[0], custom_gauss.instructions[0])

    def test_pulse_gates(self):
        """Test scheduling calibrated pulse gates."""
        q = QuantumRegister(2)
        qc = QuantumCircuit(q)
        qc.append(U2Gate(0, 0), [q[0]])
        qc.barrier(q[0], q[1])
        qc.append(U2Gate(0, 0), [q[1]])
        with self.assertWarns(DeprecationWarning):
            qc.add_calibration(
                "u2", [0], Schedule(Play(Gaussian(28, 0.2, 4), DriveChannel(0))), [0, 0]
            )
            qc.add_calibration(
                "u2", [1], Schedule(Play(Gaussian(28, 0.2, 4), DriveChannel(1))), [0, 0]
            )

        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend)
            expected = Schedule(
                Play(Gaussian(28, 0.2, 4), DriveChannel(0)),
                (28, Schedule(Play(Gaussian(28, 0.2, 4), DriveChannel(1)))),
            )
        self.assertEqual(sched.instructions, expected.instructions)

    def test_calibrated_measurements(self):
        """Test scheduling calibrated measurements."""
        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc = QuantumCircuit(q, c)
        qc.sx(0)
        qc.measure(q[0], c[0])

        with self.assertWarns(DeprecationWarning):
            meas_sched = Play(
                GaussianSquare(
                    duration=1472,
                    sigma=64,
                    width=1216,
                    amp=0.2400000000002,
                    angle=-0.247301694,
                    name="my_custom_calibration",
                ),
                MeasureChannel(0),
            )
            meas_sched |= Acquire(1472, AcquireChannel(0), MemorySlot(0))
            qc.add_calibration("measure", [0], meas_sched)

            sched = schedule(qc, self.backend)
            expected = Schedule(self.inst_map.get("sx", [0]), (16, meas_sched))
        self.assertEqual(sched.instructions, expected.instructions)

    def test_subset_calibrated_measurements(self):
        """Test that measurement calibrations can be added and used for some qubits, even
        if the other qubits do not also have calibrated measurements."""
        qc = QuantumCircuit(3, 3)
        qc.measure(0, 0)
        qc.measure(1, 1)
        qc.measure(2, 2)
        meas_scheds = []
        with self.assertWarns(DeprecationWarning):
            for qubit in [0, 2]:
                meas = Play(Gaussian(1200, 0.2, 4), MeasureChannel(qubit)) + Acquire(
                    1200, AcquireChannel(qubit), MemorySlot(qubit)
                )
                meas_scheds.append(meas)
                with self.assertWarns(DeprecationWarning):
                    qc.add_calibration("measure", [qubit], meas)

            meas = macros.measure(qubits=[1], backend=self.backend, qubit_mem_slots={0: 0, 1: 1})
            meas = meas.exclude(channels=[AcquireChannel(0), AcquireChannel(2)])
            sched = schedule(qc, self.backend)
            expected = Schedule(meas_scheds[0], meas_scheds[1], meas)
        self.assertEqual(sched.instructions, expected.instructions)

    def test_clbits_of_calibrated_measurements(self):
        """Test that calibrated measurements are only used when the classical bits also match."""
        q = QuantumRegister(2)
        c = ClassicalRegister(2)
        qc = QuantumCircuit(q, c)
        qc.measure(q[0], c[1])

        with self.assertWarns(DeprecationWarning):
            meas_sched = Play(Gaussian(1200, 0.2, 4), MeasureChannel(0))
            meas_sched |= Acquire(1200, AcquireChannel(0), MemorySlot(0))
            qc.add_calibration("measure", [0], meas_sched)

            sched = schedule(qc, self.backend)
            # Doesn't use the calibrated schedule because the classical memory slots do not match
            expected = Schedule(macros.measure([0], self.backend, qubit_mem_slots={0: 1}))
        self.assertEqual(sched.instructions, expected.instructions)

    def test_metadata_is_preserved_alap(self):
        """Test that circuit metadata is preserved in output schedule with alap."""
        q = QuantumRegister(2)
        qc = QuantumCircuit(q)
        qc.sx(q[0])
        qc.barrier(q[0], q[1])
        qc.sx(q[1])
        qc.metadata = {"experiment_type": "gst", "execution_number": "1234"}
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="alap")
        self.assertEqual({"experiment_type": "gst", "execution_number": "1234"}, sched.metadata)

    def test_metadata_is_preserved_asap(self):
        """Test that circuit metadata is preserved in output schedule with asap."""
        q = QuantumRegister(2)
        qc = QuantumCircuit(q)
        qc.sx(q[0])
        qc.barrier(q[0], q[1])
        qc.sx(q[1])
        qc.metadata = {"experiment_type": "gst", "execution_number": "1234"}
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, method="asap")
        self.assertEqual({"experiment_type": "gst", "execution_number": "1234"}, sched.metadata)

    def test_scheduler_with_params_bound(self):
        """Test scheduler with parameters defined and bound"""
        x = Parameter("x")
        qc = QuantumCircuit(2)
        qc.append(Gate("pulse_gate", 1, [x]), [0])
        with self.assertWarns(DeprecationWarning):
            expected_schedule = Schedule()
            qc.add_calibration(
                gate="pulse_gate", qubits=[0], schedule=expected_schedule, params=[x]
            )
        qc = qc.assign_parameters({x: 1})
        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend)
        self.assertEqual(sched, expected_schedule)

    def test_scheduler_with_params_not_bound(self):
        """Test scheduler with parameters defined but not bound"""
        x = Parameter("amp")
        qc = QuantumCircuit(2)
        qc.append(Gate("pulse_gate", 1, [x]), [0])
        with self.assertWarns(DeprecationWarning):
            with build() as expected_schedule:
                play(Gaussian(duration=160, amp=x, sigma=40), DriveChannel(0))
            qc.add_calibration(
                gate="pulse_gate", qubits=[0], schedule=expected_schedule, params=[x]
            )
            sched = schedule(qc, self.backend)
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(sched, transforms.target_qobj_transform(expected_schedule))

    def test_schedule_block_in_instmap(self):
        """Test schedule block in instmap can be scheduled."""
        duration = Parameter("duration")

        with self.assertWarns(DeprecationWarning):
            with build() as pulse_prog:
                play(Gaussian(duration, 0.1, 10), DriveChannel(0))

            instmap = InstructionScheduleMap()
        instmap.add("block_gate", (0,), pulse_prog, ["duration"])

        qc = QuantumCircuit(1)
        qc.append(Gate("block_gate", 1, [duration]), [0])
        qc.assign_parameters({duration: 100}, inplace=True)

        with self.assertWarns(DeprecationWarning):
            sched = schedule(qc, self.backend, inst_map=instmap)

            ref_sched = Schedule()
            ref_sched += Play(Gaussian(100, 0.1, 10), DriveChannel(0))

        self.assertEqual(sched, ref_sched)

    def test_inst_sched_map_get_measure_0(self):
        """Test that Schedule returned by backend.instruction_schedule_map.get('measure', [0])
        is actually Schedule for just qubit_0"""
        with self.assertWarns(DeprecationWarning):
            sched_from_backend = self.backend.instruction_schedule_map.get("measure", [0])
            expected_sched = Schedule(
                (
                    0,
                    Play(
                        Waveform(samples=self.pulse_2_samples, name="pulse_2"),
                        MeasureChannel(0),
                        name="pulse_2",
                    ),
                ),
                (
                    0,
                    Play(
                        Waveform(samples=self.pulse_2_samples, name="pulse_2"),
                        MeasureChannel(1),
                        name="pulse_2",
                    ),
                ),
                (
                    0,
                    Play(
                        Waveform(samples=self.pulse_2_samples, name="pulse_2"),
                        MeasureChannel(2),
                        name="pulse_2",
                    ),
                ),
                (0, Acquire(1792, AcquireChannel(0), MemorySlot(0))),
                (0, Acquire(1792, AcquireChannel(1), MemorySlot(1))),
                (0, Acquire(1792, AcquireChannel(2), MemorySlot(2))),
                name="measure",
            )
        self.assertEqual(sched_from_backend, expected_sched)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
================================================
Quantum Information (:mod:`qiskit.quantum_info`)
================================================

.. currentmodule:: qiskit.quantum_info

.. _quantum_info_operators:

Operators
=========

.. autosummary::
   :toctree: ../stubs/

   Operator
   Pauli
   Clifford
   ScalarOp
   SparseObservable
   SparsePauliOp
   CNOTDihedral
   PauliList
   pauli_basis

.. _quantum_info_states:

States
======

.. autosummary::
   :toctree: ../stubs/

   Statevector
   DensityMatrix
   StabilizerState

Channels
========
.. autosummary::
   :toctree: ../stubs/

   Choi
   SuperOp
   Kraus
   Stinespring
   Chi
   PTM

Measures
========

.. autofunction:: average_gate_fidelity
.. autofunction:: process_fidelity
.. autofunction:: gate_error
.. autofunction:: diamond_norm
.. autofunction:: state_fidelity
.. autofunction:: purity
.. autofunction:: concurrence
.. autofunction:: entropy
.. autofunction:: entanglement_of_formation
.. autofunction:: mutual_information

Utility Functions
=================

.. autosummary::
   :toctree: ../stubs/

   Quaternion

.. autofunction:: partial_trace
.. autofunction:: schmidt_decomposition
.. autofunction:: shannon_entropy
.. autofunction:: commutator
.. autofunction:: anti_commutator
.. autofunction:: double_commutator

Random
======

.. autofunction:: random_statevector
.. autofunction:: random_density_matrix
.. autofunction:: random_unitary
.. autofunction:: random_hermitian
.. autofunction:: random_pauli
.. autofunction:: random_clifford
.. autofunction:: random_quantum_channel
.. autofunction:: random_cnotdihedral
.. autofunction:: random_pauli_list

Analysis
=========

.. autofunction:: hellinger_distance
.. autofunction:: hellinger_fidelity

.. autosummary::
   :toctree: ../stubs/

   Z2Symmetries

"""

from __future__ import annotations

from qiskit._accelerate.sparse_observable import SparseObservable

from .analysis import hellinger_distance, hellinger_fidelity, Z2Symmetries
from .operators import (
    Clifford,
    Operator,
    Pauli,
    PauliList,
    ScalarOp,
    SparsePauliOp,
    anti_commutator,
    commutator,
    double_commutator,
    pauli_basis,
)
from .operators.channel import PTM, Chi, Choi, Kraus, Stinespring, SuperOp
from .operators.dihedral import CNOTDihedral
from .operators.measures import average_gate_fidelity, diamond_norm, gate_error, process_fidelity
from .random import (
    random_clifford,
    random_cnotdihedral,
    random_density_matrix,
    random_hermitian,
    random_pauli,
    random_pauli_list,
    random_quantum_channel,
    random_statevector,
    random_unitary,
)
from .states import (
    DensityMatrix,
    StabilizerState,
    Statevector,
    concurrence,
    entanglement_of_formation,
    entropy,
    mutual_information,
    partial_trace,
    purity,
    schmidt_decomposition,
    shannon_entropy,
    state_fidelity,
    negativity,
)
from .quaternion import Quaternion



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Circuit synthesis for a QFT circuit.
"""

import numpy as np
from qiskit.circuit import QuantumCircuit
from qiskit.synthesis.permutation.permutation_reverse_lnn import _append_reverse_permutation_lnn_kms
from .qft_decompose_full import _warn_if_precision_loss


def synth_qft_line(
    num_qubits: int, do_swaps: bool = True, approximation_degree: int = 0
) -> QuantumCircuit:
    """Construct a circuit for the Quantum Fourier Transform using linear
    neighbor connectivity.

    The construction is based on Fig 2.b in Fowler et al. [1].

    .. note::

        With the default value of ``do_swaps = True``, this synthesis algorithm creates a
        circuit that faithfully implements the QFT operation. When ``do_swaps = False``,
        this synthesis algorithm creates a circuit that corresponds to "QFT-with-reversal":
        applying the QFT and reversing the order of its output qubits.

    Args:
        num_qubits: The number of qubits on which the Quantum Fourier Transform acts.
        approximation_degree: The degree of approximation (0 for no approximation).
            It is possible to implement the QFT approximately by ignoring
            controlled-phase rotations with the angle beneath a threshold. This is discussed
            in more detail in https://arxiv.org/abs/quant-ph/9601018 or
            https://arxiv.org/abs/quant-ph/0403071.
        do_swaps: Whether to synthesize the "QFT" or the "QFT-with-reversal" operation.

    Returns:
        A circuit implementing the QFT operation.

    References:
        1. A. G. Fowler, S. J. Devitt, and L. C. L. Hollenberg,
           *Implementation of Shor's algorithm on a linear nearest neighbour qubit array*,
           Quantum Info. Comput. 4, 4 (July 2004), 237–251.
           `arXiv:quant-ph/0402196 [quant-ph] <https://arxiv.org/abs/quant-ph/0402196>`_
    """
    _warn_if_precision_loss(num_qubits - approximation_degree - 1)
    qc = QuantumCircuit(num_qubits)

    for i in range(num_qubits):
        qc.h(num_qubits - 1)

        for j in range(i, num_qubits - 1):
            if j - i + 2 < num_qubits - approximation_degree + 1:
                qc.p(np.pi / 2 ** (j - i + 2), num_qubits - j + i - 1)
                qc.cx(num_qubits - j + i - 1, num_qubits - j + i - 2)
                qc.p(-np.pi / 2 ** (j - i + 2), num_qubits - j + i - 2)
                qc.cx(num_qubits - j + i - 2, num_qubits - j + i - 1)
                qc.cx(num_qubits - j + i - 1, num_qubits - j + i - 2)
                qc.p(np.pi / 2 ** (j - i + 2), num_qubits - j + i - 1)
            else:
                qc.cx(num_qubits - j + i - 1, num_qubits - j + i - 2)
                qc.cx(num_qubits - j + i - 2, num_qubits - j + i - 1)
                qc.cx(num_qubits - j + i - 1, num_qubits - j + i - 2)

    if not do_swaps:
        # Add a reversal network for LNN connectivity in depth 2*n+2,
        # based on Kutin at al., https://arxiv.org/abs/quant-ph/0701194, Section 5.
        _append_reverse_permutation_lnn_kms(qc, num_qubits)

    return qc



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""T and Tdg gate."""
import math
from math import pi
from typing import Optional

import numpy

from qiskit.circuit.singleton import SingletonGate, stdlib_singleton_key
from qiskit.circuit.library.standard_gates.p import PhaseGate
from qiskit.circuit.quantumregister import QuantumRegister
from qiskit.circuit._utils import with_gate_array
from qiskit._accelerate.circuit import StandardGate


@with_gate_array([[1, 0], [0, (1 + 1j) / math.sqrt(2)]])
class TGate(SingletonGate):
    r"""Single qubit T gate (Z**0.25).

    It induces a :math:`\pi/4` phase, and is sometimes called the pi/8 gate
    (because of how the RZ(\pi/4) matrix looks like).

    This is a non-Clifford gate and a fourth-root of Pauli-Z.

    Can be applied to a :class:`~qiskit.circuit.QuantumCircuit`
    with the :meth:`~qiskit.circuit.QuantumCircuit.t` method.

    **Matrix Representation:**

    .. math::

        T = \begin{pmatrix}
                1 & 0 \\
                0 & e^{i\pi/4}
            \end{pmatrix}

    **Circuit symbol:**

    .. code-block:: text

             ┌───┐
        q_0: ┤ T ├
             └───┘

    Equivalent to a :math:`\pi/4` radian rotation about the Z axis.
    """

    _standard_gate = StandardGate.TGate

    def __init__(self, label: Optional[str] = None, *, duration=None, unit="dt"):
        """Create new T gate."""
        super().__init__("t", 1, [], label=label, duration=duration, unit=unit)

    _singleton_lookup_key = stdlib_singleton_key()

    def _define(self):
        """
        gate t a { u1(pi/4) a; }
        """
        # pylint: disable=cyclic-import
        from qiskit.circuit.quantumcircuit import QuantumCircuit

        from .u1 import U1Gate

        q = QuantumRegister(1, "q")
        qc = QuantumCircuit(q, name=self.name)
        rules = [(U1Gate(pi / 4), [q[0]], [])]
        for instr, qargs, cargs in rules:
            qc._append(instr, qargs, cargs)

        self.definition = qc

    def inverse(self, annotated: bool = False):
        """Return inverse T gate (i.e. Tdg).

        Args:
            annotated: when set to ``True``, this is typically used to return an
                :class:`.AnnotatedOperation` with an inverse modifier set instead of a concrete
                :class:`.Gate`. However, for this class this argument is ignored as the inverse
                of this gate is always a :class:`.TdgGate`.

        Returns:
            TdgGate: inverse of :class:`.TGate`
        """
        return TdgGate()

    def power(self, exponent: float, annotated: bool = False):
        return PhaseGate(0.25 * numpy.pi * exponent)

    def __eq__(self, other):
        return isinstance(other, TGate)


@with_gate_array([[1, 0], [0, (1 - 1j) / math.sqrt(2)]])
class TdgGate(SingletonGate):
    r"""Single qubit T-adjoint gate (~Z**0.25).

    It induces a :math:`-\pi/4` phase.

    This is a non-Clifford gate and a fourth-root of Pauli-Z.

    Can be applied to a :class:`~qiskit.circuit.QuantumCircuit`
    with the :meth:`~qiskit.circuit.QuantumCircuit.tdg` method.

    **Matrix Representation:**

    .. math::

        Tdg = \begin{pmatrix}
                1 & 0 \\
                0 & e^{-i\pi/4}
            \end{pmatrix}

    **Circuit symbol:**

    .. code-block:: text

             ┌─────┐
        q_0: ┤ Tdg ├
             └─────┘

    Equivalent to a :math:`-\pi/4` radian rotation about the Z axis.
    """

    _standard_gate = StandardGate.TdgGate

    def __init__(self, label: Optional[str] = None, *, duration=None, unit="dt"):
        """Create new Tdg gate."""
        super().__init__("tdg", 1, [], label=label, duration=duration, unit=unit)

    _singleton_lookup_key = stdlib_singleton_key()

    def _define(self):
        """
        gate tdg a { u1(pi/4) a; }
        """
        # pylint: disable=cyclic-import
        from qiskit.circuit.quantumcircuit import QuantumCircuit

        from .u1 import U1Gate

        q = QuantumRegister(1, "q")
        qc = QuantumCircuit(q, name=self.name)
        rules = [(U1Gate(-pi / 4), [q[0]], [])]
        for instr, qargs, cargs in rules:
            qc._append(instr, qargs, cargs)

        self.definition = qc

    def inverse(self, annotated: bool = False):
        """Return inverse Tdg gate (i.e. T).

        Args:
            annotated: when set to ``True``, this is typically used to return an
                :class:`.AnnotatedOperation` with an inverse modifier set instead of a concrete
                :class:`.Gate`. However, for this class this argument is ignored as the inverse
                of this gate is always a :class:`.TGate`.

        Returns:
            TGate: inverse of :class:`.TdgGate`
        """
        return TGate()

    def power(self, exponent: float, annotated: bool = False):
        return PhaseGate(-0.25 * numpy.pi * exponent)

    def __eq__(self, other):
        return isinstance(other, TdgGate)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Equivalence tests for quantum channel methods."""

import unittest

import numpy as np

from qiskit.quantum_info.operators.operator import Operator
from qiskit.quantum_info.operators.channel.choi import Choi
from qiskit.quantum_info.operators.channel.superop import SuperOp
from qiskit.quantum_info.operators.channel.kraus import Kraus
from qiskit.quantum_info.operators.channel.stinespring import Stinespring
from qiskit.quantum_info.operators.channel.ptm import PTM
from qiskit.quantum_info.operators.channel.chi import Chi
from .channel_test_case import ChannelTestCase


class TestEquivalence(ChannelTestCase):
    """Tests for channel equivalence.

    Tests that performing conjugate, transpose, adjoint
    in the Operator representation is equivalent to performing the same
    operations in other representations.
    """

    unitaries = [
        ChannelTestCase.UI,
        ChannelTestCase.UX,
        ChannelTestCase.UY,
        ChannelTestCase.UZ,
        ChannelTestCase.UH,
    ]

    chois = [
        ChannelTestCase.choiI,
        ChannelTestCase.choiX,
        ChannelTestCase.choiY,
        ChannelTestCase.choiZ,
        ChannelTestCase.choiH,
    ]

    chis = [
        ChannelTestCase.chiI,
        ChannelTestCase.chiX,
        ChannelTestCase.chiY,
        ChannelTestCase.chiZ,
        ChannelTestCase.chiH,
    ]

    sops = [
        ChannelTestCase.sopI,
        ChannelTestCase.sopX,
        ChannelTestCase.sopY,
        ChannelTestCase.sopZ,
        ChannelTestCase.sopH,
    ]

    ptms = [
        ChannelTestCase.ptmI,
        ChannelTestCase.ptmX,
        ChannelTestCase.ptmY,
        ChannelTestCase.ptmZ,
        ChannelTestCase.ptmH,
    ]

    def _compare_transpose_to_operator(self, chans, mats):
        """Test transpose is equivalent"""
        unitaries = [Operator(np.transpose(i)) for i in mats]
        channels = [i.transpose() for i in chans]
        for chan, uni in zip(channels, unitaries):
            self.assertEqual(chan, chan.__class__(uni))

    def _compare_conjugate_to_operator(self, chans, mats):
        """Test conjugate is equivalent"""
        unitaries = [Operator(np.conjugate(i)) for i in mats]
        channels = [i.conjugate() for i in chans]
        for chan, uni in zip(channels, unitaries):
            self.assertEqual(chan, chan.__class__(uni))

    def _compare_adjoint_to_operator(self, chans, mats):
        """Test adjoint is equivalent"""
        unitaries = [Operator(np.conjugate(np.transpose(i))) for i in mats]
        channels = [i.adjoint() for i in chans]
        for chan, uni in zip(channels, unitaries):
            self.assertEqual(chan, chan.__class__(uni))

    def test_choi_conjugate(self):
        """Test conjugate of Choi matrices is correct."""
        mats = self.unitaries
        chans = [Choi(mat) for mat in self.chois]
        self._compare_conjugate_to_operator(chans, mats)

    def test_superop_conjugate(self):
        """Test conjugate of SuperOp matrices is correct."""
        mats = self.unitaries
        chans = [SuperOp(mat) for mat in self.sops]
        self._compare_conjugate_to_operator(chans, mats)

    def test_kraus_conjugate(self):
        """Test conjugate of Kraus matrices is correct."""
        mats = self.unitaries
        chans = [Kraus(mat) for mat in mats]
        self._compare_conjugate_to_operator(chans, mats)

    def test_stinespring_conjugate(self):
        """Test conjugate of Stinespring matrices is correct."""
        mats = self.unitaries
        chans = [Stinespring(mat) for mat in mats]
        self._compare_conjugate_to_operator(chans, mats)

    def test_chi_conjugate(self):
        """Test conjugate of Chi matrices is correct."""
        mats = self.unitaries
        chans = [Chi(mat) for mat in self.chis]
        self._compare_conjugate_to_operator(chans, mats)

    def test_ptm_conjugate(self):
        """Test conjugate of PTM matrices is correct."""
        mats = self.unitaries
        chans = [PTM(mat) for mat in self.ptms]
        self._compare_conjugate_to_operator(chans, mats)

    def test_choi_conjugate_random(self):
        """Test conjugate of Choi matrices is correct."""
        mats = [self.rand_matrix(4, 4) for _ in range(4)]
        chans = [Choi(Operator(mat)) for mat in mats]
        self._compare_conjugate_to_operator(chans, mats)

    def test_superop_conjugate_random(self):
        """Test conjugate of SuperOp matrices is correct."""
        mats = [self.rand_matrix(4, 4) for _ in range(4)]
        chans = [SuperOp(Operator(mat)) for mat in mats]
        self._compare_conjugate_to_operator(chans, mats)

    def test_kraus_conjugate_random(self):
        """Test conjugate of Kraus matrices is correct."""
        mats = [self.rand_matrix(4, 4) for _ in range(4)]
        chans = [Kraus(Operator(mat)) for mat in mats]
        self._compare_conjugate_to_operator(chans, mats)

    def test_stinespring_conjugate_random(self):
        """Test conjugate of Stinespring matrices is correct."""
        mats = [self.rand_matrix(4, 4) for _ in range(4)]
        chans = [Stinespring(Operator(mat)) for mat in mats]
        self._compare_conjugate_to_operator(chans, mats)

    def test_chi_conjugate_random(self):
        """Test conjugate of Chi matrices is correct."""
        mats = [self.rand_matrix(4, 4) for _ in range(4)]
        chans = [Chi(Operator(mat)) for mat in mats]
        self._compare_conjugate_to_operator(chans, mats)

    def test_ptm_conjugate_random(self):
        """Test conjugate of PTM matrices is correct."""
        mats = [self.rand_matrix(4, 4) for _ in range(4)]
        chans = [PTM(Operator(mat)) for mat in mats]
        self._compare_conjugate_to_operator(chans, mats)

    def test_choi_transpose(self):
        """Test transpose of Choi matrices is correct."""
        mats = self.unitaries
        chans = [Choi(mat) for mat in self.chois]
        self._compare_transpose_to_operator(chans, mats)

    def test_superop_transpose(self):
        """Test transpose of SuperOp matrices is correct."""
        mats = self.unitaries
        chans = [SuperOp(mat) for mat in self.sops]
        self._compare_transpose_to_operator(chans, mats)

    def test_kraus_transpose(self):
        """Test transpose of Kraus matrices is correct."""
        mats = self.unitaries
        chans = [Kraus(mat) for mat in mats]
        self._compare_transpose_to_operator(chans, mats)

    def test_stinespring_transpose(self):
        """Test transpose of Stinespring matrices is correct."""
        mats = self.unitaries
        chans = [Stinespring(mat) for mat in mats]
        self._compare_transpose_to_operator(chans, mats)

    def test_chi_transpose(self):
        """Test transpose of Chi matrices is correct."""
        mats = self.unitaries
        chans = [Chi(mat) for mat in self.chis]
        self._compare_transpose_to_operator(chans, mats)

    def test_ptm_transpose(self):
        """Test transpose of PTM matrices is correct."""
        mats = self.unitaries
        chans = [PTM(mat) for mat in self.ptms]
        self._compare_transpose_to_operator(chans, mats)

    def test_choi_transpose_random(self):
        """Test transpose of Choi matrices is correct."""
        mats = [self.rand_matrix(4, 4) for _ in range(4)]
        chans = [Choi(Operator(mat)) for mat in mats]
        self._compare_transpose_to_operator(chans, mats)

    def test_superop_transpose_random(self):
        """Test transpose of SuperOp matrices is correct."""
        mats = [self.rand_matrix(4, 4) for _ in range(4)]
        chans = [SuperOp(Operator(mat)) for mat in mats]
        self._compare_transpose_to_operator(chans, mats)

    def test_kraus_transpose_random(self):
        """Test transpose of Kraus matrices is correct."""
        mats = [self.rand_matrix(4, 4) for _ in range(4)]
        chans = [Kraus(Operator(mat)) for mat in mats]
        self._compare_transpose_to_operator(chans, mats)

    def test_stinespring_transpose_random(self):
        """Test transpose of Stinespring matrices is correct."""
        mats = [self.rand_matrix(4, 4) for _ in range(4)]
        chans = [Stinespring(Operator(mat)) for mat in mats]
        self._compare_transpose_to_operator(chans, mats)

    def test_chi_transpose_random(self):
        """Test transpose of Chi matrices is correct."""
        mats = [self.rand_matrix(4, 4) for _ in range(4)]
        chans = [Chi(Operator(mat)) for mat in mats]
        self._compare_transpose_to_operator(chans, mats)

    def test_ptm_transpose_random(self):
        """Test transpose of PTM matrices is correct."""
        mats = [self.rand_matrix(4, 4) for _ in range(4)]
        chans = [PTM(Operator(mat)) for mat in mats]
        self._compare_transpose_to_operator(chans, mats)

    def test_choi_adjoint(self):
        """Test adjoint of Choi matrices is correct."""
        mats = self.unitaries
        chans = [Choi(mat) for mat in self.chois]
        self._compare_adjoint_to_operator(chans, mats)

    def test_superop_adjoint(self):
        """Test adjoint of SuperOp matrices is correct."""
        mats = self.unitaries
        chans = [SuperOp(mat) for mat in self.sops]
        self._compare_adjoint_to_operator(chans, mats)

    def test_kraus_adjoint(self):
        """Test adjoint of Kraus matrices is correct."""
        mats = self.unitaries
        chans = [Kraus(mat) for mat in mats]
        self._compare_adjoint_to_operator(chans, mats)

    def test_stinespring_adjoint(self):
        """Test adjoint of Stinespring matrices is correct."""
        mats = self.unitaries
        chans = [Stinespring(mat) for mat in mats]
        self._compare_adjoint_to_operator(chans, mats)

    def test_chi_adjoint(self):
        """Test adjoint of Chi matrices is correct."""
        mats = self.unitaries
        chans = [Chi(mat) for mat in self.chis]
        self._compare_adjoint_to_operator(chans, mats)

    def test_ptm_adjoint(self):
        """Test adjoint of PTM matrices is correct."""
        mats = self.unitaries
        chans = [PTM(mat) for mat in self.ptms]
        self._compare_adjoint_to_operator(chans, mats)

    def test_choi_adjoint_random(self):
        """Test adjoint of Choi matrices is correct."""
        mats = [self.rand_matrix(4, 4) for _ in range(4)]
        chans = [Choi(Operator(mat)) for mat in mats]
        self._compare_adjoint_to_operator(chans, mats)

    def test_superop_adjoint_random(self):
        """Test adjoint of SuperOp matrices is correct."""
        mats = [self.rand_matrix(4, 4) for _ in range(4)]
        chans = [SuperOp(Operator(mat)) for mat in mats]
        self._compare_adjoint_to_operator(chans, mats)

    def test_kraus_adjoint_random(self):
        """Test adjoint of Kraus matrices is correct."""
        mats = [self.rand_matrix(4, 4) for _ in range(4)]
        chans = [Kraus(Operator(mat)) for mat in mats]
        self._compare_adjoint_to_operator(chans, mats)

    def test_stinespring_adjoint_random(self):
        """Test adjoint of Stinespring matrices is correct."""
        mats = [self.rand_matrix(4, 4) for _ in range(4)]
        chans = [Stinespring(Operator(mat)) for mat in mats]
        self._compare_adjoint_to_operator(chans, mats)

    def test_chi_adjoint_random(self):
        """Test adjoint of Chi matrices is correct."""
        mats = [self.rand_matrix(4, 4) for _ in range(4)]
        chans = [Chi(Operator(mat)) for mat in mats]
        self._compare_adjoint_to_operator(chans, mats)

    def test_ptm_adjoint_random(self):
        """Test adjoint of PTM matrices is correct."""
        mats = [self.rand_matrix(4, 4) for _ in range(4)]
        chans = [PTM(Operator(mat)) for mat in mats]
        self._compare_adjoint_to_operator(chans, mats)


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test the EnlargeWithAncilla pass"""

import unittest

from qiskit import QuantumRegister, QuantumCircuit
from qiskit.transpiler import Layout
from qiskit.transpiler.passes import EnlargeWithAncilla
from qiskit.converters import circuit_to_dag
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class TestEnlargeWithAncilla(QiskitTestCase):
    """Tests the EnlargeWithAncilla pass."""

    def setUp(self):
        super().setUp()
        self.qr3 = QuantumRegister(3, "qr")
        circuit = QuantumCircuit(self.qr3)
        circuit.h(self.qr3)
        self.dag = circuit_to_dag(circuit)

    def test_no_extension(self):
        """There are no virtual qubits to extend."""
        layout = Layout({self.qr3[0]: 0, self.qr3[1]: 1, self.qr3[2]: 2})

        pass_ = EnlargeWithAncilla()
        pass_.property_set["layout"] = layout
        after = pass_.run(self.dag)

        qregs = list(after.qregs.values())
        self.assertEqual(1, len(qregs))
        self.assertEqual(self.qr3, qregs[0])

    def test_with_extension(self):
        """There are 2 virtual qubit to extend."""
        ancilla = QuantumRegister(2, "ancilla")

        layout = Layout(
            {0: self.qr3[0], 1: ancilla[0], 2: self.qr3[1], 3: ancilla[1], 4: self.qr3[2]}
        )
        layout.add_register(ancilla)

        pass_ = EnlargeWithAncilla()
        pass_.property_set["layout"] = layout
        after = pass_.run(self.dag)

        qregs = list(after.qregs.values())
        self.assertEqual(2, len(qregs))
        self.assertEqual(self.qr3, qregs[0])
        self.assertEqual(ancilla, qregs[1])


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for core modules of pulse drawer."""

from qiskit import pulse
from qiskit.visualization.pulse_v2 import drawings, types
from test import QiskitTestCase  # pylint: disable=wrong-import-order
from qiskit.utils.deprecate_pulse import decorate_test_methods, ignore_pulse_deprecation_warnings


@decorate_test_methods(ignore_pulse_deprecation_warnings)
class TestDrawingObjects(QiskitTestCase):
    """Tests for DrawingObjects."""

    @ignore_pulse_deprecation_warnings
    def setUp(self) -> None:
        """Setup."""
        super().setUp()

        # bits
        self.ch_d0 = [pulse.DriveChannel(0)]
        self.ch_d1 = [pulse.DriveChannel(1)]

        # metadata
        self.meta1 = {"val1": 0, "val2": 1}
        self.meta2 = {"val1": 2, "val2": 3}

        # style data
        self.style1 = {"property1": 0, "property2": 1}
        self.style2 = {"property1": 2, "property2": 3}

    def test_line_data_equivalent(self):
        """Test for LineData."""
        xs = [0, 1, 2]
        ys = [3, 4, 5]

        data1 = drawings.LineData(
            data_type="test",
            xvals=xs,
            yvals=ys,
            channels=self.ch_d0,
            meta=self.meta1,
            ignore_scaling=True,
            styles=self.style1,
        )

        data2 = drawings.LineData(
            data_type="test",
            xvals=xs,
            yvals=ys,
            channels=self.ch_d1,
            meta=self.meta2,
            ignore_scaling=True,
            styles=self.style2,
        )

        self.assertEqual(data1, data2)

    def test_line_data_equivalent_with_abstract_coordinate(self):
        """Test for LineData with abstract coordinate."""
        xs = [types.AbstractCoordinate.LEFT, types.AbstractCoordinate.RIGHT]
        ys = [types.AbstractCoordinate.TOP, types.AbstractCoordinate.BOTTOM]

        data1 = drawings.LineData(
            data_type="test",
            xvals=xs,
            yvals=ys,
            channels=self.ch_d0,
            meta=self.meta1,
            ignore_scaling=True,
            styles=self.style1,
        )

        data2 = drawings.LineData(
            data_type="test",
            xvals=xs,
            yvals=ys,
            channels=self.ch_d1,
            meta=self.meta2,
            ignore_scaling=True,
            styles=self.style2,
        )

        self.assertEqual(data1, data2)

    def test_text_data(self):
        """Test for TextData."""
        xs = [0]
        ys = [1]

        data1 = drawings.TextData(
            data_type="test",
            xvals=xs,
            yvals=ys,
            text="test1",
            latex=r"test_1",
            channels=self.ch_d0,
            meta=self.meta1,
            ignore_scaling=True,
            styles=self.style1,
        )

        data2 = drawings.TextData(
            data_type="test",
            xvals=xs,
            yvals=ys,
            text="test2",
            latex=r"test_2",
            channels=self.ch_d1,
            meta=self.meta2,
            ignore_scaling=True,
            styles=self.style2,
        )

        self.assertEqual(data1, data2)

    def test_text_data_with_abstract_coordinate(self):
        """Test for TextData with abstract coordinates."""
        xs = [types.AbstractCoordinate.RIGHT]
        ys = [types.AbstractCoordinate.TOP]

        data1 = drawings.TextData(
            data_type="test",
            xvals=xs,
            yvals=ys,
            text="test1",
            latex=r"test_1",
            channels=self.ch_d0,
            meta=self.meta1,
            ignore_scaling=True,
            styles=self.style1,
        )

        data2 = drawings.TextData(
            data_type="test",
            xvals=xs,
            yvals=ys,
            text="test2",
            latex=r"test_2",
            channels=self.ch_d1,
            meta=self.meta2,
            ignore_scaling=True,
            styles=self.style2,
        )

        self.assertEqual(data1, data2)

    def test_box_data(self):
        """Test for BoxData."""
        xs = [0, 1]
        ys = [-1, 1]

        data1 = drawings.BoxData(
            data_type="test",
            xvals=xs,
            yvals=ys,
            channels=self.ch_d0,
            meta=self.meta1,
            ignore_scaling=True,
            styles=self.style1,
        )

        data2 = drawings.BoxData(
            data_type="test",
            xvals=xs,
            yvals=ys,
            channels=self.ch_d1,
            meta=self.meta2,
            ignore_scaling=True,
            styles=self.style2,
        )

        self.assertEqual(data1, data2)

    def test_box_data_with_abstract_coordinate(self):
        """Test for BoxData with abstract coordinate."""
        xs = [types.AbstractCoordinate.LEFT, types.AbstractCoordinate.RIGHT]
        ys = [types.AbstractCoordinate.BOTTOM, types.AbstractCoordinate.TOP]

        data1 = drawings.BoxData(
            data_type="test",
            xvals=xs,
            yvals=ys,
            channels=self.ch_d0,
            meta=self.meta1,
            ignore_scaling=True,
            styles=self.style1,
        )

        data2 = drawings.BoxData(
            data_type="test",
            xvals=xs,
            yvals=ys,
            channels=self.ch_d1,
            meta=self.meta2,
            ignore_scaling=True,
            styles=self.style2,
        )

        self.assertEqual(data1, data2)
